# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import abc

import six

from .constants import OP


@six.add_metaclass(abc.ABCMeta)
class Operator(object):
    def __init__(self, op):
        self.op = op

    @abc.abstractmethod
    def expr(self):
        pass

    def __repr__(self):
        return "operator:{}".format(self.op)


@six.add_metaclass(abc.ABCMeta)
class LogicalOperator(Operator):
    def __init__(self, op, content):
        super(LogicalOperator, self).__init__(op)

        # TODO: valid the content, should be list
        self.content = content

    def expr(self):
        separator = " %s " % self.op
        return "(%s)" % separator.join([c.expr() for c in self.content])

    def render(self, obj_set):
        separator = " %s " % self.op
        return "(%s)" % separator.join([c.render(obj_set) for c in self.content])

    @abc.abstractmethod
    def eval(self, obj_set):
        pass

    # TODO: expr with values, should change
    # @abc.abstractmethod
    # def expr_translate()


class AndOperator(LogicalOperator):
    def __init__(self, content):
        super(AndOperator, self).__init__(OP.AND, content)

    def eval(self, obj_set):
        # return all([c.eval(obj_set) for c in self.content])
        # Short-circuit evaluation
        for c in self.content:
            if not c.eval(obj_set):
                return False
        return True


class OrOperator(LogicalOperator):
    def __init__(self, content):
        super(OrOperator, self).__init__(OP.OR, content)

    def eval(self, obj_set):
        # return any([c.eval(obj_set) for c in self.content])
        # Short-circuit evaluation
        for c in self.content:
            if c.eval(obj_set):
                return True
        return False


@six.add_metaclass(abc.ABCMeta)
class BinaryOperator(Operator):
    def __init__(self, op, field, value):
        super(BinaryOperator, self).__init__(op)
        self.field = field
        self.value = value

    def render(self, obj_set):
        value = self.value
        if isinstance(self.value, six.string_types):
            value = "'%s'" % self.value

        attr = obj_set.get(self.field)
        if isinstance(attr, six.string_types):
            attr = "'%s'" % attr

        return "(%s %s %s)" % (attr, self.op, value)

    def expr(self):
        value = self.value
        if isinstance(self.value, six.string_types):
            value = "'%s'" % self.value

        return "(%s %s %s)" % (self.field, self.op, value)

    @abc.abstractmethod
    def calculate(self, left, right):
        pass

    def _eval_positive(self, attr, attr_is_array, value, value_is_array):  # NOQA
        """
        positive:
        - 1   hit: return True
        - all miss: return False

        e.g.
        op = eq => one of attr equals one of value

        attr = 1; value = 1; True
        attr = 1; value = [1, 2]; True
        attr = [1, 2]; value = 2; True
        attr = [1, 2]; value = [5, 1]; True

        attr = [1, 2]; value = [3, 4]; False
        """
        if self.op == OP.ANY:
            return self.calculate(attr, value)

        # 1. IN/NOT_IN value is a list, just only check attr
        if self.op in (OP.IN,):
            if attr_is_array:
                for a in attr:
                    if self.calculate(a, value):
                        return True
                return False

            return self.calculate(attr, value)

        # 2. CONTAINS/NOT_CONTAINS attr is a list, just check value
        if self.op in (OP.CONTAINS,):
            if value_is_array:
                for v in value:
                    if self.calculate(attr, v):
                        return True
                return False

            return self.calculate(attr, value)

        # 3. Others, check both attr and value
        # 3.1 both not array, the most common situation
        if not (value_is_array or attr_is_array):
            return self.calculate(attr, value)

        # 3.2 only value is array, the second common situation
        if value_is_array and (not attr_is_array):
            for v in value:
                # return early if hit
                if self.calculate(attr, v):
                    return True
            return False

        # 3.3 only attr value is array
        if (not value_is_array) and attr_is_array:
            for a in attr:
                # return early if hit
                if self.calculate(a, value):
                    return True
            return False

        # 4. both array
        for a in attr:
            for v in value:
                # return early if hit
                if self.calculate(a, v):
                    return True
        return False

    def _eval_negative(self, attr, attr_is_array, value, value_is_array):  # NOQA
        """
        negative:
        - 1   miss: return False
        - all hit: return True

        e.g.
        op = not_eq => all of attr should not_eq to all of the value

        attr = 1; value = 2; True
        attr = 1; value = [2]; True
        attr = [1, 2]; value = [3, 4]; True
        attr = [1, 2]; value = 3; True

        attr = [1, 2]; value = [2, 3]; False
        """
        # 1. IN/NOT_IN value is a list, just only check attr
        if self.op in (OP.NOT_IN,):
            if attr_is_array:
                for a in attr:
                    if not self.calculate(a, value):
                        return False
                return True

            return self.calculate(attr, value)

        # 2. CONTAINS/NOT_CONTAINS attr is a list, just check value
        if self.op in (OP.NOT_CONTAINS,):
            if value_is_array:
                for v in value:
                    if not self.calculate(attr, v):
                        return False
                return True

            return self.calculate(attr, value)

        # 3. Others, check both attr and value
        # 3.1 both not array, the most common situation
        if not (value_is_array or attr_is_array):
            return self.calculate(attr, value)

        # 3.2 only value is array, the second common situation
        if value_is_array and (not attr_is_array):
            for v in value:
                if not self.calculate(attr, v):
                    return False
            return True

        # 3.3 only attr value is array
        if (not value_is_array) and attr_is_array:
            for a in attr:
                if not self.calculate(a, value):
                    return False
            return True

        # 4. both array
        for a in attr:
            for v in value:
                # return early if hit
                if not self.calculate(a, v):
                    return False
        return True

    def eval(self, obj_set):
        """
        type: str/numberic/boolean
        the value support `type` or `[type]`
        the obj_set.get(self.field) support `type` or `[type]`

        if one of them is array, or both array
        calculate each item in array
        """
        attr = obj_set.get(self.field)
        value = self.value

        attr_is_array = isinstance(attr, (list, tuple))
        value_is_array = isinstance(value, (list, tuple))

        # positive and negative operator
        # ==  命中一个即返回
        # !=  需要全部遍历完, 确认全部不等于才返回?
        if self.op.startswith("not_"):
            return self._eval_negative(attr, attr_is_array, value, value_is_array)
        else:
            return self._eval_positive(attr, attr_is_array, value, value_is_array)


class EqualOperator(BinaryOperator):
    def __init__(self, field, value):
        super(EqualOperator, self).__init__(OP.EQ, field, value)

    def calculate(self, left, right):
        return left == right


class NotEqualOperator(BinaryOperator):
    def __init__(self, field, value):
        super(NotEqualOperator, self).__init__(OP.NOT_EQ, field, value)

    def calculate(self, left, right):
        return left != right


class InOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: value should be list or string(sequence?)
        super(InOperator, self).__init__(OP.IN, field, value)

    def calculate(self, left, right):
        return left in right


class NotInOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: value should be list or string(sequence?)
        super(NotInOperator, self).__init__(OP.NOT_IN, field, value)

    def calculate(self, left, right):
        return left not in right


class ContainsOperator(BinaryOperator):
    def __init__(self, field, value):
        super(ContainsOperator, self).__init__(OP.CONTAINS, field, value)

    def calculate(self, left, right):
        return right in left


class NotContainsOperator(BinaryOperator):
    def __init__(self, field, value):
        super(NotContainsOperator, self).__init__(OP.NOT_CONTAINS, field, value)

    def calculate(self, left, right):
        return right not in left


class StartsWithOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: value should be string?
        super(StartsWithOperator, self).__init__(OP.STARTS_WITH, field, value)

    def calculate(self, left, right):
        return left.startswith(right)


class NotStartsWithOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: value should be string?
        super(NotStartsWithOperator, self).__init__(OP.NOT_STARTS_WITH, field, value)

    def calculate(self, left, right):
        return not left.startswith(right)


class EndsWithOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: value should be string?
        super(EndsWithOperator, self).__init__(OP.ENDS_WITH, field, value)

    def calculate(self, left, right):
        return left.endswith(right)


class NotEndsWithOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: value should be string?
        super(NotEndsWithOperator, self).__init__(OP.NOT_ENDS_WITH, field, value)

    def calculate(self, left, right):
        return not left.endswith(right)


class LTOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: field / value should be numberic
        super(LTOperator, self).__init__(OP.LT, field, value)

    def calculate(self, left, right):
        return left < right


class LTEOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: field / value should be numberic
        super(LTEOperator, self).__init__(OP.LTE, field, value)

    def calculate(self, left, right):
        return left <= right


class GTOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: field / value should be numberic
        super(GTOperator, self).__init__(OP.GT, field, value)

    def calculate(self, left, right):
        return left > right


class GTEOperator(BinaryOperator):
    def __init__(self, field, value):
        # TODO: field / value should be numberic
        super(GTEOperator, self).__init__(OP.GTE, field, value)

    def calculate(self, left, right):
        return left >= right


class AnyOperator(BinaryOperator):
    def __init__(self, field, value):
        super(AnyOperator, self).__init__(OP.ANY, field, value)

    def calculate(self, left, right):
        return True


BINARY_OPERATORS = {
    OP.EQ: EqualOperator,
    OP.NOT_EQ: NotEqualOperator,
    OP.IN: InOperator,
    OP.NOT_IN: NotInOperator,
    OP.CONTAINS: ContainsOperator,
    OP.NOT_CONTAINS: NotContainsOperator,
    OP.STARTS_WITH: StartsWithOperator,
    OP.NOT_STARTS_WITH: NotStartsWithOperator,
    OP.ENDS_WITH: EndsWithOperator,
    OP.NOT_ENDS_WITH: NotEndsWithOperator,
    OP.LT: LTOperator,
    OP.LTE: LTEOperator,
    OP.GT: GTOperator,
    OP.GTE: GTEOperator,
    OP.ANY: AnyOperator,
}



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import six

from iam.eval.constants import OP
from iam.eval.expression import field_value_convert

from .base import Converter

"""
convert the expression into sql where clause

FIXME:
- sql injection via value
- single quote `'` in the value string, sql syntax error
"""


class SQLConverter(Converter):
    def __init__(self, key_mapping=None):
        super(SQLConverter, self).__init__(key_mapping)

    def _to_str_present(self, value, wrap_str):
        if not wrap_str:
            return value

        # FIXME: current not support the value with single quote '
        if isinstance(value, six.string_types):
            return "'%s'" % value
        return value

    def _positive(self, fmt, left, right, wrap_str=True):
        """
        if right is an array, [1, 2, 3]
        should be
        (left = 1 OR left = 2 OR left = 3)
        (left LIKE '1%' OR left LIKE '2%' OR left LIKE '3%')
        """
        is_array = isinstance(right, (list, tuple))
        if is_array:
            condition = " OR ".join([fmt.format(left, self._to_str_present(i, wrap_str)) for i in right])
            return "({})".format(condition)

        return fmt.format(left, self._to_str_present(right, wrap_str))

    def _negative(self, fmt, left, right, wrap_str=True):
        """
        if right is an array, [1, 2, 3]
        should be
        (left != 1 AND left != 2 AND left != 3)
        (left NOT LIKE '1%' AND left NOT LIKE '2%' and left NOT LIKE '3%')
        """
        is_array = isinstance(right, (list, tuple))
        if is_array:
            condition = " AND ".join([fmt.format(left, self._to_str_present(i, wrap_str)) for i in right])
            return "({})".format(condition)

        return fmt.format(left, self._to_str_present(right, wrap_str))

    def _eq(self, left, right):
        return self._positive("{} = {}", left, right)

    def _not_eq(self, left, right):
        return self._negative("{} != {}", left, right)

    def _in(self, left, right):
        # TODO: right shuld be a list
        right = [self._to_str_present(r, True) for r in right]
        return "{} IN ({})".format(left, ",".join([str(r) for r in right]))

    def _not_in(self, left, right):
        # TODO: right shuld be a list
        right = [self._to_str_present(r, True) for r in right]
        return "{} NOT IN ({})".format(left, ",".join([str(r) for r in right]))

    def _contains(self, left, right):
        raise NotImplementedError

    def _not_contains(self, left, right):
        raise NotImplementedError

    def _starts_with(self, left, right):
        # FIXME: right with single quote, will fail
        return self._positive("{} LIKE '{}%'", left, right, False)

    def _not_starts_with(self, left, right):
        return self._negative("{} NOT LIKE '{}%'", left, right, False)

    def _ends_with(self, left, right):
        return self._positive("{} LIKE '%{}'", left, right, False)

    def _not_ends_with(self, left, right):
        return self._negative("{} NOT LIKE '%{}'", left, right, False)

    def _lt(self, left, right):
        return self._positive("{} < {}", left, right)

    def _lte(self, left, right):
        return self._positive("{} <= {}", left, right)

    def _gt(self, left, right):
        return self._positive("{} > {}", left, right)

    def _gte(self, left, right):
        return self._positive("{} >= {}", left, right)

    def _any(self, left, right):
        return "1 = 1"

    def _and(self, content):
        condition = " AND ".join([self.convert(c) for c in content])
        return "(%s)" % condition

    def _or(self, content):
        condition = " OR ".join([self.convert(c) for c in content])
        return "(%s)" % condition

    def convert(self, data):
        op = data["op"]

        if op == OP.AND:
            return self._and(data["content"])
        elif op == OP.OR:
            return self._or(data["content"])

        op_func = {
            OP.EQ: self._eq,
            OP.NOT_EQ: self._not_eq,
            OP.IN: self._in,
            OP.NOT_IN: self._not_in,
            OP.CONTAINS: self._contains,
            OP.NOT_CONTAINS: self._not_contains,
            OP.STARTS_WITH: self._starts_with,
            OP.NOT_STARTS_WITH: self._not_starts_with,
            OP.ENDS_WITH: self._ends_with,
            OP.NOT_ENDS_WITH: self._not_ends_with,
            OP.LT: self._lt,
            OP.LTE: self._lte,
            OP.GT: self._gt,
            OP.GTE: self._gte,
            OP.ANY: self._any,
        }.get(op)

        if op_func is None:
            raise ValueError("invalid op %s" % op)

        value = data["value"]
        field = data["field"]

        # 权限中心保留字预处理
        field, value = field_value_convert(op, field, value)

        if self.key_mapping and field in self.key_mapping:
            field = self.key_mapping.get(field)

        return op_func(field, value)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import codecs
import json
import os
import sys

from django.conf import settings

from iam.contrib.iam_migration import exceptions
from iam.contrib.iam_migration.utils import do_migrate


def upsert_system_render(data):
    resource_api_host = getattr(settings, "BK_IAM_RESOURCE_API_HOST", None)
    if resource_api_host:
        data["provider_config"]["host"] = resource_api_host


renders = {"upsert_system": upsert_system_render}


class IAMMigrator(object):
    def __init__(self, migration_json):
        self.migration_json = migration_json

    def migrate(self):
        iam_host = settings.BK_IAM_INNER_HOST
        app_code = settings.APP_CODE
        app_secret = settings.SECRET_KEY

        # only trigger migrator at db migrate
        if "migrate" not in sys.argv:
            return

        json_path = getattr(settings, "BK_IAM_MIGRATION_JSON_PATH", "support-files/iam/")
        file_path = os.path.join(settings.BASE_DIR, json_path, self.migration_json)

        with codecs.open(file_path, mode="r", encoding="utf-8") as fp:
            data = json.load(fp=fp)

        # data pre render
        for op in data["operations"]:
            if op["operation"] in renders:
                renders[op["operation"]](op["data"])

        ok, _ = do_migrate.api_ping(iam_host)
        if not ok:
            raise exceptions.NetworkUnreachableError("bk iam ping error")

        ok = do_migrate.do_migrate(data, iam_host, app_code, app_secret)
        if not ok:
            raise exceptions.MigrationFailError("iam migrate fail")



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import abc
import logging

import six
from tastypie.authorization import Authorization, ReadOnlyAuthorization, Unauthorized
from tastypie.exceptions import ImmediateHttpResponse

from iam import Action, Request
from iam.contrib.django.response import IAMAuthFailedResponse
from iam.exceptions import AuthFailedException

logger = logging.getLogger("iam")


@six.add_metaclass(abc.ABCMeta)
class IAMAuthorizationHelper(object):
    def __init__(
        self,
        system,
        create_action,
        read_action,
        update_action,
        delete_action,
        filter_key_mapping={},
    ):
        self.system = system
        self.create_action = create_action
        self.read_action = read_action
        self.update_action = update_action
        self.delete_action = delete_action
        self.filter_key_mapping = filter_key_mapping

    @abc.abstractmethod
    def get_subject(self, bundle):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_create_detail_resources(self, bundle):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def get_create_detail_environment(self, bundle):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_read_detail_resources(self, bundle):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def get_read_detail_environment(self, bundle):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_update_detail_resources(self, bundle):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def get_update_detail_environment(self, bundle):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_delete_detail_resources(self, bundle):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def get_delete_detail_environment(self, bundle):
        raise NotImplementedError()

    @abc.abstractclassmethod
    def get_read_list_environment(self, bundle):
        raise NotImplementedError()


class IAMCreateAuthorizationMixin(object):
    def create_list(self, object_list, bundle):
        raise Unauthorized()

    def create_detail(self, object_list, bundle):
        system = self.helper.system
        subject = self.helper.get_subject(bundle)
        action = Action(self.helper.create_action)
        resources = self.helper.get_create_detail_resources(bundle)

        request = Request(
            system,
            subject,
            action,
            resources,
            self.helper.get_create_detail_environment(bundle),
        )

        allowed = self.iam.is_allowed(request)
        logger.debug("tastypie create_detail is_allowed request({}) result: {}".format(request.to_dict(), allowed))

        if not allowed:
            raise ImmediateHttpResponse(IAMAuthFailedResponse(AuthFailedException(system, subject, action, resources)))

        return allowed


class IAMUpdateAuthorizationMixin(object):
    def update_list(self, object_list, bundle):
        raise Unauthorized()

    def update_detail(self, object_list, bundle):
        system = self.helper.system
        subject = self.helper.get_subject(bundle)
        action = Action(self.helper.update_action)
        resources = self.helper.get_update_detail_resources(bundle)

        request = Request(
            system,
            subject,
            action,
            resources,
            self.helper.get_update_detail_environment(bundle),
        )

        allowed = self.iam.is_allowed(request)
        logger.debug("tastypie update_detail is_allowed request({}) result: {}".format(request.to_dict(), allowed))

        if not allowed:
            raise ImmediateHttpResponse(IAMAuthFailedResponse(AuthFailedException(system, subject, action, resources)))

        return allowed


class IAMDeleteAuthorizationMixin(object):
    def delete_list(self, object_list, bundle):
        raise Unauthorized()

    def delete_detail(self, object_list, bundle):
        system = self.helper.system
        subject = self.helper.get_subject(bundle)
        action = Action(self.helper.delete_action)
        resources = self.helper.get_delete_detail_resources(bundle)

        request = Request(
            system,
            subject,
            action,
            resources,
            self.helper.get_delete_detail_environment(bundle),
        )

        allowed = self.iam.is_allowed(request)
        logger.debug("tastypie delete_detail is_allowed request({}) result: {}".format(request.to_dict(), allowed))

        if not allowed:
            raise ImmediateHttpResponse(IAMAuthFailedResponse(AuthFailedException(system, subject, action, resources)))

        return allowed


class IAMReadDetailAuthorizationMixin(object):
    def read_detail(self, object_list, bundle):
        system = self.helper.system
        subject = self.helper.get_subject(bundle)
        action = Action(self.helper.read_action)
        resources = self.helper.get_read_detail_resources(bundle)

        request = Request(
            system,
            subject,
            action,
            resources,
            self.helper.get_read_detail_environment(bundle),
        )

        allowed = self.iam.is_allowed(request)
        logger.debug("tastypie read_detail is_allowed request({}) result: {}".format(request.to_dict(), allowed))

        if not allowed:
            raise ImmediateHttpResponse(IAMAuthFailedResponse(AuthFailedException(system, subject, action, resources)))

        return allowed


class IAMAuthorization(Authorization):
    def __init__(self, iam, helper):
        self.iam = iam
        self.helper = helper


class IAMReadOnlyAuthorization(ReadOnlyAuthorization):
    def __init__(self, iam, helper):
        self.iam = iam
        self.helper = helper


class ReadOnlyCompleteListIAMAuthorization(
    IAMReadDetailAuthorizationMixin,
    IAMReadOnlyAuthorization,
):
    def read_list(self, object_list, bundle):
        return object_list


class CustomCreateCompleteListIAMAuthorization(
    IAMUpdateAuthorizationMixin,
    IAMDeleteAuthorizationMixin,
    IAMReadDetailAuthorizationMixin,
    IAMAuthorization,
):
    def read_list(self, object_list, bundle):
        return object_list

    def create_detail(self, object_list, bundle):
        return True


class CompleteListIAMAuthorization(
    IAMCreateAuthorizationMixin,
    IAMUpdateAuthorizationMixin,
    IAMDeleteAuthorizationMixin,
    IAMReadDetailAuthorizationMixin,
    IAMAuthorization,
):
    def read_list(self, object_list, bundle):
        return object_list


class StrictListIAMAuthorization(
    IAMCreateAuthorizationMixin,
    IAMUpdateAuthorizationMixin,
    IAMDeleteAuthorizationMixin,
    IAMReadDetailAuthorizationMixin,
    IAMAuthorization,
):
    def read_list(self, object_list, bundle):
        request = Request(
            system=self.system,
            subject=self.get_subject(bundle),
            action=Action(self.read_action),
            resources=[],
            environment=self.get_read_list_environment(bundle),
        )
        f = self.iam.make_filter(request, key_mapping=self.helper.filter_key_mapping)
        logger.debug("tastypie read_list make_filter request({}) result: {}".format(request.to_dict(), f))
        return object_list.filter(f)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


from django.db.models import Q

from .api.client import Client
from .apply.models import Application
from .auth.models import ApiAuthRequest, MultiActionRequest, Request, Resource
from .contrib.converter.queryset import DjangoQuerySetConverter
from .eval.object import ObjectSet
from .exceptions import AuthAPIError, AuthInvalidParam, AuthInvalidRequest


class DummyIAM(object):
    """
    input: object
    """

    def __init__(self, app_code, app_secret, bk_iam_host, bk_paas_host):
        self._client = Client(app_code, app_secret, bk_iam_host, bk_paas_host)

    def _do_policy_query(self, request, with_resources=True):
        return []

    def _do_policy_query_by_actions(self, request, with_resources=True):
        return []

    def _eval_expr(self, expr, obj_set):
        return True

    def _eval_policy(self, policy, obj_set):
        return True

    def _build_object_set(self, system, resources, only_local=False):
        obj_set = ObjectSet()

        resource_id_list = []

        # if no resources or resources is None
        if not resources:
            return obj_set, ""

        for resource in resources:
            # only local resource need to be calculated
            # 跨系统资源依赖的策略在服务端就计算完了, 策略表达式中只会存在本系统的
            if only_local and (resource.system != system):
                continue

            attrs = resource.attribute
            attrs["id"] = resource.id
            obj_set.add_object(resource.type, attrs)

            resource_id_list.append((resource.type, resource.id))

        # 如果只有一个本地资源, 直接返回不带类型的ID;
        # [("flow", "1")]   => "1"
        # 如果存在层级资源 返回 {type},{id}/{type2},{id2}
        # [("cluster", "a"), ("area", "b")) =>  "cluster,a/area,b"

        # NOTE: 这里不会存在=> 跨系统资源依赖=>上面已经过滤掉了
        # [("job", "script", "a"), ("cmdb", "host", "b")) =>  "job:script,a/cmdb:host,b"

        resource_id = ""
        if len(resource_id_list) == 1:
            resource_id = resource_id_list[0][1]
        else:
            resource_id = "/".join(["%s,%s" % (a[0], a[1]) for a in resource_id_list])

        return obj_set, resource_id

    def _validate_request(self, request):
        if not isinstance(request, Request):
            raise AuthInvalidRequest("request should be instance of iam.auth.models.Request")

        request.validate()

    def _validate_multi_action_request(self, request):
        if not isinstance(request, MultiActionRequest):
            raise AuthInvalidRequest("request should be instance of iam.auth.models.MultiActionRequest")

        request.validate()

    def _validate_resources_list(self, resources_list):
        # resources_list = [resources] = [[Resource], [Resource]]
        if not isinstance(resources_list, list):
            raise AuthInvalidParam("resources_list should be list of [iam.auth.models.Resource]")

        if not all([isinstance(r, list) for r in resources_list]):
            raise AuthInvalidParam("resources_list should be list of [iam.auth.models.Resource]")

        if not all([isinstance(i, Resource) for r in resources_list for i in r]):
            raise AuthInvalidParam("resources should be list of iam.auth.models.Resource")

    def _validate_resources_list_same_local_only(self, system, resources_list):
        # 校验, resources_list中只能是本地同一类的资源
        resource_types = {}
        for rs in resources_list:
            for r in rs:
                if system != r.system:
                    raise AuthInvalidParam(
                        "resources_list not support auth for resource belong other system: %s" % r.system
                    )
                resource_types[r.type] = 1

        if len(resource_types) != 1:
            raise AuthInvalidParam(
                "resources_list should all with the same resource_type, but got %s" % resource_types.keys()
            )

    def is_allowed(self, request):
        """
        单个资源是否有权限校验
        request中会带resource到IAM, IAM会进行两阶段计算, 即resources也会参与到计算中

        支持:
        - 本地资源 resources中只有本地资源
        - 跨系统资源依赖 resources中有本地也有远程资源 (此时resoruces一定要传, 因为需要IAM帮助获取跨系统资源)
        """
        return True

    def batch_is_allowed(self, request, resources_list):
        """
        多个资源是否有权限校验
        request中不会带resource到IAM, IAM不会会进行两阶段计算, 直接返回system+action+subejct的所有策略
        然后逐一计算

        - 一次策略查询, 多次计算

        支持:
        - 本地资源 resources中只有本地资源
        - **不支持**跨系统资源依赖
        """
        # NOTE: 不向服务端传任何resource
        result = {}
        for resources in resources_list:
            _, resource_id = self._build_object_set(request.system, resources, only_local=False)
            result[resource_id] = True
        return result

    def resource_multi_actions_allowed(self, request):
        """
        单个资源多个action是否有权限校验
        request中会带resource到IAM, IAM会进行两阶段计算, 即resources也会参与到计算中

        支持:
        - 本地资源 resources中只有本地资源
        - 跨系统资源依赖 resources中有本地也有远程资源 (此时resoruces一定要传, 因为需要IAM帮助获取跨系统资源)
        """
        actions_allowed = {}
        for a in request.to_dict()["actions"]:
            action = a["id"]
            actions_allowed[action] = True
        return actions_allowed

    def batch_resource_multi_actions_allowed(self, request, resources_list):
        """
        批量资源多个action是否有权限校验
        request中会带resource到IAM, IAM会进行两阶段计算, 即resources也会参与到计算中

        支持:
        - 本地资源 resources中只有本地资源
        - **不支持**跨系统资源依赖
        """
        resources_actions_perms = {}
        for resources in resources_list:
            for a in request.to_dict()["actions"]:
                action = a["id"]
                obj_set, resource_id = self._build_object_set(request.system, resources, only_local=False)
                resources_actions_perms.setdefault(resource_id, {})[action] = True
        return resources_actions_perms

    def make_filter(self, request, converter_class=DjangoQuerySetConverter, key_mapping=None):
        return ~Q(pk=None)

    # TODO: add the register model apis
    def get_token(self, system):
        # bool, message, token
        return self._client.get_token(system)

    def is_basic_auth_allowed(self, system, basic_auth):
        return True

    def get_apply_url(self, application, bk_token=None, bk_username=None):
        if isinstance(application, dict):
            data = application
        elif isinstance(application, Application):
            # do validate
            application.validate()
            data = application.to_dict()
        else:
            raise AuthInvalidRequest("application shuld be instance of dict or iam.apply.modles.Application")

        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        # bool, message, url
        return self._client.get_apply_url(bk_token, bk_username, data)

    def grant_resource_creator_actions(self, application, bk_token=None, bk_username=None):
        return True, "success"

    def grant_resource_creator_action_attributes(self, application, bk_token=None, bk_username=None):
        return True, "success"

    def grant_batch_resource_creator_actions(self, application, bk_token=None, bk_username=None):
        return True, "success"

    def grant_or_revoke_instance_permission(self, request, bk_token=None, bk_username=None):
        if not isinstance(request, ApiAuthRequest):
            raise AuthInvalidRequest("request should be a instance of iam.auth.models.ApiAuthRequest")

        self._validate_request(request)
        data = request.to_dict()

        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        ok, message, policies = self._client.instance_authorization(bk_token, bk_username, data)
        if not ok:
            raise AuthAPIError(message)
        return policies

    def grant_or_revoke_path_permission(self, request, bk_token=None, bk_username=None):
        if not isinstance(request, ApiAuthRequest):
            raise AuthInvalidRequest("request should be a instance of iam.auth.models.ApiAuthRequest")
        self._validate_request(request)
        data = request.to_dict()

        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        ok, message, policies = self._client.path_authorization(bk_token, bk_username, data)
        if not ok:
            raise AuthAPIError(message)
        return policies



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import codecs
import os
import sys
from datetime import datetime

from django.apps import apps
from django.conf import settings
from django.core.management.base import BaseCommand
from django.db.migrations.loader import MigrationLoader
from django.template.engine import Context, Engine

from iam.contrib.iam_migration import conf
from iam.contrib.iam_migration.apps import IAMMigrationConfig
from iam.contrib.iam_migration.template import migration_template


class Command(BaseCommand):

    help = "Create new migration for specific iam migration json file e.g. python manage.py iam_makemigrations migration.json"

    def add_arguments(self, parser):
        parser.add_argument("migration_json", nargs="?", type=str)

    def handle(self, *args, **options):
        json_file = options["migration_json"]
        if not json_file:
            sys.stderr.write("please provide a migration json file name\n")
            exit(1)

        json_path = getattr(settings, "BK_IAM_MIGRATION_JSON_PATH", "support-files/iam/")
        file_path = os.path.join(settings.BASE_DIR, json_path, json_file)

        if not os.path.exists(file_path):
            sys.stderr.write("{} is not exist.\n".format(file_path))
            exit(1)

        loader = MigrationLoader(None, ignore_no_migrations=False)
        migration_leaf = loader.graph.leaf_nodes(conf.MIGRATION_APP_NAME)

        is_initial = True
        last_migration_name = None

        if migration_leaf:
            is_initial = False
            last_migration_name = migration_leaf[0][1]

        sys.stdout.write("is initial migration: {}\n".format(is_initial))
        sys.stdout.write("last migration: {}\n".format(last_migration_name))

        migration_name = self.migration_name(last_migration_name)
        migration_file = "{}.py".format(
            os.path.join(apps.get_app_config(conf.MIGRATION_APP_NAME).path, "migrations", migration_name)
        )

        with codecs.open(migration_file, mode="w", encoding="utf-8") as fp:
            template = Engine.get_default().from_string(migration_template)
            fp.write(
                template.render(
                    Context(
                        {
                            "migration_json": json_file,
                            "app_label": IAMMigrationConfig.name,
                            "initial": is_initial,
                            "last_migration_name": last_migration_name,
                        }
                    )
                )
            )

    def migration_name(self, last_migration_name):
        system_id = getattr(settings, "BK_IAM_SYSTEM_ID", None)
        time = datetime.now().strftime("%Y%m%d%H%M")

        if not system_id:
            self.stderr.write("You must set BK_IAM_SYSTEM_ID in django settings before make migrations")
            exit(1)

        if not last_migration_name:
            return "0001_initial"

        code = "%04d" % (int(last_migration_name[:4]) + 1)

        return "{code}_{system_id}_{time}".format(code=code, system_id=system_id, time=time)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import abc

import six

"""
build the application

# 无资源实例的操作权限, 例如 访问开发者中心
{
  "system_id": "bk_job",  # 权限的系统
  "actions": [
    {
      "id": "execute_job",  # 操作id
      "related_resource_types": []
    }
  ]
}

# 有资源实例的操作权限, 例如 管理应用framework
{
  "system_id": "bk_job",  # 权限的系统
  "actions": [
    {
      "id": "execute_job",  # 操作id
      "related_resource_types": [  # 关联的资源类型, 无关联资源类型的操作, 可以为空
        {
          "system_id": "bk_job",  # 资源类型所属的系统id
          "type": "job",  # 资源类型
          "instances": [  # 申请权限的资源实例
            [  # 带层级的实例表示
              {
                "type": "job",  # 层级节点的资源类型
                "id": "job1",  # 层级节点的资源实例id
                "name": "作业1"  # 层级节点的资源名称
              }
            ]
          ]
        },
        {
          "system_id": "bk_cmdb",  # 资源类型所属的系统id
          "type": "host",  # 操作依赖的另外一个资源类型
          "instances": [
            [
              {
                "type": "biz",
                "id": "biz1",
                "name": "业务1"
              }, {
                "type": "set",
                "id": "set1",
                "name": "集群1"
              }, {
                "type": "module",
                "id": "module1",
                "name": "模块1"
              }, {
                "type": "host",
                "id": "host1",
                "name": "主机1"
              }
            ]
          ]
        }
      ]
    }
  ],
  "environment": {  # 权限的可用条件
    "source_system_id": ""  # 选填, 来源系统, 如果需要限制权限的来源系统可填
  }
}
"""
# =========== resource


class ResourceNode(object):
    def __init__(self, type, id, name):
        self.type = type
        self.id = id
        self.name = name

    def to_dict(self):
        return {
            "type": self.type,
            "id": self.id,
            "name": self.name,
        }


class ResourceInstance(object):
    def __init__(self, resource_nodes):
        # should not be empty and should be type of ResourceNode
        self.resource_nodes = resource_nodes

    def validate(self):
        if not self.resource_nodes:
            raise ValueError(
                "ResourceInstance.resource_nodes invalid: should contain at least 1 iam.apply.models.ResourceNode"
            )

        if not isinstance(self.resource_nodes, list):
            raise TypeError("ResourceInstance.resource_nodes should be a list of iam.apply.models.ResourceNode")

        for i, r in enumerate(self.resource_nodes):
            if not isinstance(r, ResourceNode):
                raise ValueError("ResourceInstance.resource_instances[%d] should be instance of ResourceNode" % i)

    def to_dict(self):
        return [n.to_dict() for n in self.resource_nodes]


class RelatedResourceType(object):
    def __init__(self, system_id, type, instances):
        self.system_id = system_id
        self.type = type
        # should not be empty and should be type of Resource
        self.instances = instances

    def validate(self):
        if not isinstance(self.instances, list):
            raise TypeError("ResourceInstance.instances should be a list of iam.apply.models.ResourceInstance")

        for i, r in enumerate(self.instances):
            try:
                r.validate()
            except Exception as e:
                raise ValueError("RelatedResourceType.instances[%d] invalid: %s" % (i, e))

    def to_dict(self):
        return {"system_id": self.system_id, "type": self.type, "instances": [r.to_dict() for r in self.instances]}


# =========== action
@six.add_metaclass(abc.ABCMeta)
class BaseAction(object):
    def __init__(self, id):
        self.id = id

    @abc.abstractmethod
    def validate(self):
        pass

    @abc.abstractmethod
    def to_dict(self):
        pass

    def __repr__(self):
        return "action:{}".format(self.id)


class ActionWithoutResources(BaseAction):
    def __init__(self, id):
        super(ActionWithoutResources, self).__init__(id)

    def validate(self):
        pass

    def to_dict(self):
        return {"id": self.id, "related_resource_types": []}


class ActionWithResources(BaseAction):
    def __init__(self, id, related_resource_types):
        super(ActionWithResources, self).__init__(id)
        # TODO: should not be empty, a list of ResourceType
        self.related_resource_types = related_resource_types

    def validate(self):
        if not self.related_resource_types:
            raise ValueError("Action.related_resource_types invalid: should contain at least 1 RelatedResourceType")

        if not isinstance(self.related_resource_types, list):
            raise TypeError("Action.related_resource_types should be a list of iam.apply.models.RelatedResourceType")

        for i, rt in enumerate(self.related_resource_types):
            if not isinstance(rt, RelatedResourceType):
                raise TypeError(
                    "Action.related_resource_types[%d] should be instance of " "iam.apply.model.RelatedResourceType" % i
                )
            try:
                rt.validate()
            except Exception as e:
                raise ValueError("Action.related_resource_types[%d] invalid: %s" % (i, e))

    def to_dict(self):
        return {"id": self.id, "related_resource_types": [r.to_dict() for r in self.related_resource_types]}


# =========== application
class Application(object):
    def __init__(self, system_id, actions, environment=None):
        self.system_id = system_id
        self.actions = actions
        # self.environment = environment or {}

    def to_dict(self):
        return {
            "system_id": self.system_id,
            "actions": [a.to_dict() for a in self.actions],
            # "environment": self.environment
        }

    def validate(self):
        if not self.actions:
            raise ValueError("Application.actions invalid: should contain at least 1 Action")

        if not isinstance(self.actions, list):
            raise TypeError("Application.actions should be a list of iam.apply.models.Action")

        for i, action in enumerate(self.actions):
            if not isinstance(action, BaseAction):
                raise TypeError(
                    "Application.actions[%d] should be instance of "
                    "iam.apply.models.ActionWithResources/ActionWithoutResources" % i
                )
            try:
                action.validate()
            except Exception as e:
                raise ValueError("Application.actions[%d] invalid: %s" % (i, e))



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import abc

import six


class ListResult(object):
    def __init__(self, results, count):
        """
        :param results: 返回的结果
        :param count: 总记录数
        """
        self.count = count
        self.results = results

    def to_dict(self):
        return {"count": self.count, "results": self.results}

    def to_list(self):
        return self.results


@six.add_metaclass(abc.ABCMeta)
class ResourceProvider(object):
    @abc.abstractmethod
    def list_attr(self, **options):
        """
        处理来自 iam 的 list_attr 请求
        return: ListResult
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def list_attr_value(self, filter, page, **options):
        """
        处理来自 iam 的 list_attr_value 请求
        return: ListResult
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def list_instance(self, filter, page, **options):
        """
        处理来自 iam 的 list_instance 请求
        return: ListResult
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def fetch_instance_info(self, filter, **options):
        """
        处理来自 iam 的 fetch_instance_info 请求
        return: ListResult
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def list_instance_by_policy(self, filter, page, **options):
        """
        处理来自 iam 的 list_instance_by_policy 请求
        return: ListResult
        """
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


# Set default logging handler to avoid "No handler found" warnings.
import logging
from logging import NullHandler

from .__version__ import __version__  # noqa
from .auth.models import Action, MultiActionRequest, Request, Resource, Subject  # noqa
from .contrib.converter.base import Converter  # noqa
from .contrib.converter.queryset import DjangoQuerySetConverter, PathEqDjangoQuerySetConverter  # noqa
from .contrib.converter.sql import SQLConverter  # noqa
from .dummy_iam import DummyIAM  # noqa
from .eval.constants import OP  # noqa
from .eval.expression import make_expression  # noqa
from .eval.object import DictObject, ObjectSet  # noqa
from .eval.operators import (  # noqa
    AndOperator,
    AnyOperator,
    BinaryOperator,
    ContainsOperator,
    EndsWithOperator,
    EqualOperator,
    GTEOperator,
    GTOperator,
    InOperator,
    LogicalOperator,
    LTEOperator,
    LTOperator,
    NotContainsOperator,
    NotEndsWithOperator,
    NotEqualOperator,
    NotInOperator,
    NotStartsWithOperator,
    Operator,
    OrOperator,
    StartsWithOperator,
)
from .iam import IAM  # noqa

logging.getLogger(__name__).addHandler(NullHandler())



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


try:
    import ujson as json
except Exception:
    import json

import logging

from django.http.response import JsonResponse
from django.views.decorators.csrf import csrf_exempt

from iam.contrib.django.dispatcher.exceptions import InvalidPageException, KeywordTooShortException
from iam.exceptions import AuthInvalidOperation
from iam.resource.dispatcher import ResourceApiDispatcher
from iam.resource.provider import ResourceProvider
from iam.resource.utils import get_filter_obj, get_page_obj

logger = logging.getLogger("iam")


def fail_response(code, message, request_id):
    response = JsonResponse({"code": code, "result": False, "message": message, "data": None})
    response["X-Request-Id"] = request_id
    return response


def success_response(data, request_id):
    response = JsonResponse({"code": 0, "result": True, "message": "", "data": data})
    response["X-Request-Id"] = request_id
    return response


class DjangoBasicResourceApiDispatcher(ResourceApiDispatcher):
    def __init__(self, iam, system):
        self.iam = iam
        self.system = system
        self._provider = {}

    def register(self, provider_type, provider):
        if not issubclass(type(provider), ResourceProvider):
            raise AuthInvalidOperation("provider must be subclass of iam.resource.provider.ResourceProvider")

        if provider_type in self._provider:
            raise AuthInvalidOperation("provider {} already been registered".format(provider_type))

        self._provider[provider_type] = provider

    def as_view(self, decorators=[]):
        @csrf_exempt
        def view(request):
            return self._dispatch(request)

        for dec in decorators:
            view = dec(view)

        return view

    def _dispatch(self, request):

        request_id = request.META.get("HTTP_X_REQUEST_ID", "")

        # auth check
        auth = request.META.get("HTTP_AUTHORIZATION", "")
        auth_allowed = self.iam.is_basic_auth_allowed(self.system, auth)

        if not auth_allowed:
            logger.error("resource request({}) auth failed with auth param: {}".format(request_id, auth))
            return fail_response(401, "basic auth failed", request_id)

        # load json data
        try:
            data = json.loads(request.body)
        except Exception:
            logger.error("resource request({}) failed with invalid body: {}".format(request_id, request.body))
            return fail_response(400, "reqeust body is not a valid json", request_id)

        # check basic params
        method = data.get("method")
        resource_type = data.get("type")
        if not (method and resource_type):
            logger.error("resource request({}) failed with invalid data: {}".format(request_id, data))
            return fail_response(400, "method and type is required field", request_id)

        # check resource type
        if resource_type not in self._provider:
            logger.error(
                "resource request({}) failed with unsupport resource type: {}".format(request_id, resource_type)
            )
            return fail_response(404, "unsupport resource type: {}".format(resource_type), request_id)

        # check method and process
        processor = getattr(self, "_dispatch_{}".format(method), None)
        if not processor:
            logger.error("resource request({}) failed with unsupport method: {}".format(request_id, method))
            return fail_response(404, "unsupport method: {}".format(method), request_id)

        logger.info(
            "resource request({}) with filter: {}, page: {}".format(request_id, data.get("filter"), data.get("page"))
        )
        try:
            return processor(request, data, request_id)
        except InvalidPageException as e:
            return fail_response(422, str(e), request_id)
        except KeywordTooShortException as e:
            return fail_response(406, str(e), request_id)
        except Exception as e:
            logger.exception("resource request({}) failed with exception: {}".format(request_id, e))
            return fail_response(500, str(e), request_id)

    def _get_options(self, request):
        return {"language": request.META.get("HTTP_BLUEKING_LANGUAGE", "zh-cn")}

    def _dispatch_list_attr(self, request, data, request_id):
        options = self._get_options(request)

        provider = self._provider[data["type"]]

        pre_process = getattr(provider, "pre_list_attr", None)
        if pre_process and callable(pre_process):
            pre_process(**options)

        result = provider.list_attr(**options)

        return success_response(result.to_list(), request_id)

    def _dispatch_list_attr_value(self, request, data, request_id):
        options = self._get_options(request)

        filter_obj = get_filter_obj(data.get("filter"), ["attr", "keyword", "ids"])
        page_obj = get_page_obj(data.get("page"))

        provider = self._provider[data["type"]]

        pre_process = getattr(provider, "pre_list_attr_value", None)
        if pre_process and callable(pre_process):
            pre_process(filter_obj, page_obj, **options)

        result = provider.list_attr_value(filter_obj, page_obj, **options)

        return success_response(result.to_dict(), request_id)

    def _dispatch_list_instance(self, request, data, request_id):
        options = self._get_options(request)

        filter_obj = get_filter_obj(data.get("filter"), ["parent", "search", "resource_type_chain"])
        page_obj = get_page_obj(data.get("page"))

        provider = self._provider[data["type"]]

        pre_process = getattr(provider, "pre_list_instance", None)
        if pre_process and callable(pre_process):
            pre_process(filter_obj, page_obj, **options)

        result = provider.list_instance(filter_obj, page_obj, **options)

        return success_response(result.to_dict(), request_id)

    def _dispatch_fetch_instance_info(self, request, data, request_id):
        options = self._get_options(request)

        filter_obj = get_filter_obj(data.get("filter"), ["ids", "attrs"])

        provider = self._provider[data["type"]]

        pre_process = getattr(provider, "pre_fetch_instance_info", None)
        if pre_process and callable(pre_process):
            pre_process(filter_obj, **options)

        result = provider.fetch_instance_info(filter_obj, **options)

        return success_response(result.to_list(), request_id)

    def _dispatch_list_instance_by_policy(self, request, data, request_id):
        options = self._get_options(request)

        filter_obj = get_filter_obj(data.get("filter"), ["expression"])
        page_obj = get_page_obj(data.get("page"))

        provider = self._provider[data["type"]]

        pre_process = getattr(provider, "pre_list_instance_by_policy", None)
        if pre_process and callable(pre_process):
            pre_process(filter_obj, page_obj, **options)

        result = provider.list_instance_by_policy(filter_obj, page_obj, **options)

        return success_response(result.to_list(), request_id)

    def _dispatch_search_instance(self, request, data, request_id):
        options = self._get_options(request)

        filter_obj = get_filter_obj(data.get("filter"), ["parent", "keyword"])

        if filter_obj.keyword is None or len(filter_obj.keyword) < 2:
            raise KeywordTooShortException("the length of keyword should be greater than or equals to 2")

        page_obj = get_page_obj(data.get("page"))

        provider = self._provider[data["type"]]

        pre_process = getattr(provider, "pre_search_instance", None)
        if pre_process and callable(pre_process):
            pre_process(filter_obj, page_obj, **options)

        search_function = getattr(provider, "search_instance", None)
        if not (search_function and callable(search_function)):
            return fail_response(404, "resource type: {} not support search instance".format(data["type"]), request_id)

        result = provider.search_instance(filter_obj, page_obj, **options)

        return success_response(result.to_dict(), request_id)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import abc

import six


@six.add_metaclass(abc.ABCMeta)
class Converter(object):
    def __init__(self, key_mapping=None):
        self.key_mapping = key_mapping

    @abc.abstractmethod
    def _eq(self, left, right):
        pass

    @abc.abstractmethod
    def _not_eq(self, left, right):
        pass

    @abc.abstractmethod
    def _in(self, left, right):
        pass

    @abc.abstractmethod
    def _not_in(self, left, right):
        pass

    @abc.abstractmethod
    def _contains(self, left, right):
        pass

    @abc.abstractmethod
    def _not_contains(self, left, right):
        pass

    @abc.abstractmethod
    def _starts_with(self, left, right):
        pass

    @abc.abstractmethod
    def _not_starts_with(self, left, right):
        pass

    @abc.abstractmethod
    def _ends_with(self, left, right):
        pass

    @abc.abstractmethod
    def _not_ends_with(self, left, right):
        pass

    @abc.abstractmethod
    def _lt(self, left, right):
        pass

    @abc.abstractmethod
    def _lte(self, left, right):
        pass

    @abc.abstractmethod
    def _gt(self, left, right):
        pass

    @abc.abstractmethod
    def _gte(self, left, right):
        pass

    @abc.abstractmethod
    def _any(self, left, right):
        pass

    @abc.abstractmethod
    def _and(self, content):
        pass

    @abc.abstractmethod
    def _or(self, content):
        pass

    @abc.abstractmethod
    def convert(self, data):
        pass



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


from __future__ import unicode_literals

import json
import logging
import os
import time

from cachetools import TTLCache, cached
from requests.models import PreparedRequest

from .http import http_delete, http_get, http_post, http_put

logger = logging.getLogger("iam")

BK_IAM_VERSION = "1"


class Client(object):
    """
    input: json
    """

    def __init__(self, app_code, app_secret, bk_iam_host, bk_paas_host):
        self._app_code = app_code
        self._app_secret = app_secret
        self._host = bk_iam_host
        self._bk_paas_host = bk_paas_host

        # will add ?debug=true in url, for debug api/policy, show the details
        isApiDebugEnabled = os.environ.get("IAM_API_DEBUG") == "true" or os.environ.get("BKAPP_IAM_API_DEBUG") == "true"
        # will add ?force=true in url, for api/policy run without cache(all data from database)
        isApiForceEnabled = os.environ.get("IAM_API_FORCE") == "true" or os.environ.get("BKAPP_IAM_API_FORCE") == "true"

        self._extra_url_params = {}
        if isApiDebugEnabled:
            self._extra_url_params["debug"] = "true"
        if isApiForceEnabled:
            self._extra_url_params["force"] = "true"

    def _call_api(self, http_func, host, path, data, headers, timeout=None):
        url = "{host}{path}".format(host=host, path=path)

        begin = time.time()

        # add extra params in url if not empty
        if self._extra_url_params:
            preReq = PreparedRequest()
            preReq.prepare_url(url, self._extra_url_params)
            url = preReq.url

        ok, message, _data = http_func(url, data, headers=headers, timeout=timeout)

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("do http request: method=`%s`, url=`%s`, data=`%s`", http_func.__name__, url, json.dumps(data))
            logger.debug("http request result: ok=`%s`, message=`%s`, _data=`%s`", ok, message, json.dumps(_data))
            logger.debug("http request took %s ms", int((time.time() - begin) * 1000))

        if not ok:
            return False, message or "verify from iam server fail", None

        if _data.get("code") != 0:
            return False, _data.get("message") or "iam api fail", None

        _d = _data.get("data")

        return True, "ok", _d

    def _call_iam_api(self, http_func, path, data, timeout=None):
        headers = {
            "X-BK-APP-CODE": self._app_code,
            "X-BK-APP-SECRET": self._app_secret,
            "X-Bk-IAM-Version": BK_IAM_VERSION,
        }
        return self._call_api(http_func, self._host, path, data, headers, timeout=timeout)

    def _call_esb_api(self, http_func, path, data, bk_token, bk_username, timeout=None):
        headers = {
            # "BK-APP-CODE": self._app_code,
            # "BK-APP-SECRET": self._app_secret,
        }
        data.update(
            {
                "bk_app_code": self._app_code,
                "bk_app_secret": self._app_secret,
                "bk_token": bk_token,
                "bk_username": bk_username,
            }
        )
        return self._call_api(http_func, self._bk_paas_host, path, data, headers, timeout=timeout)

    # ---------- system
    def add_system(self, data):
        # data.id required
        path = "/api/v1/model/systems"
        ok, message, data = self._call_iam_api(http_post, path, data)
        # if alreay exists, return true
        return ok, message

    def update_system(self, system_id, data):
        # data.id required
        path = "/api/v1/model/systems/{system_id}".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    # ---------- resource_type
    def batch_add_resource_types(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/resource-types".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        # if alreay exists, return true
        return ok, message

    def update_resource_type(self, system_id, resource_type_id, data):
        path = ("/api/v1/model/systems/{system_id}/resource-types/{resource_type_id}").format(
            system_id=system_id, resource_type_id=resource_type_id
        )

        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    def batch_delete_resource_types(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/resource-types?check_existence=false".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_delete, path, data)
        return ok, message

    # ---------- action
    def batch_add_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message

    def update_action(self, system_id, action_id, data):
        path = "/api/v1/model/systems/{system_id}/actions/{action_id}".format(system_id=system_id, action_id=action_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    def batch_delete_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/actions?check_existence=false".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_delete, path, data)
        return ok, message

    # register create association permission action.
    def add_resource_creator_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/resource_creator_actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message

    # update create association permission action.
    def update_resource_creator_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/resource_creator_actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    # return resource instance creator to iam, esb needed.
    def grant_resource_creator_actions(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/authorization/resource_creator_action/"

        ok, message, _data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message

        return True, "success"

    # return resource instance creator action attribute to iam, esb needed.
    def grant_resource_creator_action_attributes(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/authorization/resource_creator_action_attribute/"

        ok, message, data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message

        return True, "success"

    # return resource instance creator to iam, esb needed.
    def grant_batch_resource_creator_actions(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/authorization/batch_resource_creator_action/"

        ok, message, _data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message

        return True, "success"

    # ---------- action-topology
    def add_action_topology(self, system_id, action_type, data):
        path = "/api/v1/model/systems/{system_id}/action-topologies/{action_type}".format(
            system_id=system_id, action_type=action_type
        )
        ok, message, data = self._call_iam_api(http_post, path, data)
        # if alreay exists, return true
        return ok, message

    def update_action_topology(self, system_id, action_type, data):
        path = "/api/v1/model/systems/{system_id}/action-topologies/{action_type}".format(
            system_id=system_id, action_type=action_type
        )
        ok, message, data = self._call_iam_api(http_put, path, data)
        # if alreay exists, return true
        return ok, message

    # ---------- query
    def query(self, system_id):
        path = "/api/v1/model/systems/{system_id}/query".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_get, path, None)
        return ok, message, data

    # ---------- ping
    def ping(self):
        url = "{host}{path}".format(host=self._host, path="/ping")
        ok, _, data = http_get(url, None, timeout=5)
        return ok, data

    # ---------- query system_id_set/resource_type_id_set, action_id_set
    @cached(cache=TTLCache(maxsize=10, ttl=30))
    def query_all_models(self, system_id):
        ok, message, data = self.query(system_id)
        if not ok:
            return set(), set(), set()

        system = data.get("base_info", {})
        resource_types = data.get("resource_types", [])
        actions = data.get("actions", [])

        system_id_set = {system.get("id")}
        resource_type_id_set = {r.get("id") for r in resource_types}
        action_id_set = {a.get("id") for a in actions}

        return system_id_set, resource_type_id_set, action_id_set

    # --------- upset
    def upsert_system(self, system_id, data):
        system_id_set, _, _ = self.query_all_models(system_id)

        if system_id not in system_id_set:
            return self.add_system(system_id, data)
        return self.update_system(system_id, data)

    def upsert_resource_type(self, system_id, data):
        # data.id required
        d_resource_type_id = data.get("id")
        if not d_resource_type_id:
            return False, "the field `id` required"

        _, resource_id_set, _ = self.query_all_models(system_id)

        if d_resource_type_id not in resource_id_set:
            return self.add_resource_type(system_id, data)
        return self.update_resource_type(system_id, data)

    def upsert_action(self, system_id, data):
        d_action_id = data.get("id")
        if not d_action_id:
            return False, "the field `id` required"

        _, _, action_id_set = self.query_all_models(system_id)

        if d_action_id not in action_id_set:
            return self.add_action(system_id, data)
        return self.update_action(system_id, data)

    # --------- policy
    def policy_query(self, data):
        path = "/api/v1/policy/query"
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message, data

    def policy_query_by_actions(self, data):
        path = "/api/v1/policy/query_by_actions"
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message, data

    def get_token(self, system_id):
        path = "/api/v1/model/systems/{system_id}/token".format(system_id=system_id)
        ok, message, _data = self._call_iam_api(http_get, path, {})
        if not ok:
            return False, message, ""

        return True, "success", _data.get("token", "")

    # apply
    def get_apply_url(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/application/"

        ok, message, _data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message, ""

        return True, "success", _data.get("url", "")

    def instance_authorization(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/authorization/instance/"
        ok, message, _data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message, ""
        return True, "success", _data.get("token", "")

    def batch_instance_authorization(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/authorization/batch_instance/"
        ok, message, _data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message, ""
        return True, "success", _data

    def path_authorization(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/authorization/path/"
        ok, message, _data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message, ""
        return True, "success", _data.get("token", "")

    def batch_path_authorization(self, bk_token, bk_username, data):
        path = "/api/c/compapi/v2/iam/authorization/batch_path/"
        ok, message, _data = self._call_esb_api(http_post, path, data, bk_token, bk_username, timeout=5)
        if not ok:
            return False, message, ""
        return True, "success", _data

    def query_policies_with_action_id(self, system_id, data):
        path = "/api/v1/systems/{system_id}/policies".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_get, path, data)
        if not ok:
            return False, message, ""
        return True, message, data



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import base64
import logging
import sys
import time
from copy import deepcopy

from cachetools import TTLCache, cached
from six import string_types

from .api.client import Client
from .apply.models import Application
from .auth.models import ApiAuthRequest, ApiBatchAuthRequest, MultiActionRequest, Request, Resource
from .cache import hash_key
from .contrib.converter.queryset import DjangoQuerySetConverter
from .eval.expression import make_expression
from .eval.object import ObjectSet
from .exceptions import AuthAPIError, AuthInvalidParam, AuthInvalidRequest

logger = logging.getLogger("iam")


class IAM(object):
    """
    input: object
    """

    def __init__(self, app_code, app_secret, bk_iam_host, bk_paas_host):
        self._client = Client(app_code, app_secret, bk_iam_host, bk_paas_host)

    def _do_policy_query(self, request, with_resources=True):
        data = request.to_dict()
        logger.debug("the request: %s", data)

        # NOTE: 不向服务端传任何resource, 用于统一类资源的批量鉴权
        # 将会返回所有策略, 然后遍历资源列表和策略列表, 逐一计算
        if not with_resources:
            data["resources"] = []

        ok, message, policies = self._client.policy_query(data)
        if not ok:
            raise AuthAPIError(message)
        return policies

    @cached(cache=TTLCache(maxsize=1024, ttl=60), key=hash_key)
    def _do_policy_query_with_cache(self, request):
        return self._do_policy_query(request)

    def _do_policy_query_by_actions(self, request, with_resources=True):
        data = request.to_dict()
        logger.debug("the request: %s", data)

        # NOTE: 不向服务端传任何resource, 用于统一类资源的批量鉴权
        # 将会返回所有策略, 然后遍历资源列表和策略列表, 逐一计算
        if not with_resources:
            data["resources"] = []

        ok, message, action_policies = self._client.policy_query_by_actions(data)
        if not ok:
            raise AuthAPIError(message)
        return action_policies

    def _eval_expr(self, expr, obj_set):
        logger.debug("the return expr: %s", expr.expr())
        logger.debug("the return expr render: %s", expr.render(obj_set))

        # 5. eval and return
        eval_begin = time.time()
        allowed = expr.eval(obj_set)
        logger.debug("the return expr eval: %s", allowed)
        eval_time = int((time.time() - eval_begin) * 1000)
        logger.debug("the return expr eval took %s ms", eval_time)
        return allowed

    def _eval_policy(self, policy, obj_set):
        if not policy:
            return False

        expr = make_expression(policy)

        return self._eval_expr(expr, obj_set)

    def _build_object_set(self, system, resources, only_local=False):
        obj_set = ObjectSet()

        resource_id_list = []

        # if no resources or resources is None
        if not resources:
            return obj_set, ""

        for resource in resources:
            # only local resource need to be calculated
            # 跨系统资源依赖的策略在服务端就计算完了, 策略表达式中只会存在本系统的
            if only_local and (resource.system != system):
                continue

            attrs = resource.attribute
            attrs["id"] = resource.id
            obj_set.add_object(resource.type, attrs)

            resource_id_list.append((resource.type, resource.id))

        # 如果只有一个本地资源, 直接返回不带类型的ID;
        # [("flow", "1")]   => "1"
        # 如果存在层级资源 返回 {type},{id}/{type2},{id2}
        # [("cluster", "a"), ("area", "b")) =>  "cluster,a/area,b"

        # NOTE: 这里不会存在=> 跨系统资源依赖=>上面已经过滤掉了
        # [("job", "script", "a"), ("cmdb", "host", "b")) =>  "job:script,a/cmdb:host,b"

        resource_id = ""
        if len(resource_id_list) == 1:
            resource_id = resource_id_list[0][1]
        else:
            resource_id = "/".join(["%s,%s" % (a[0], a[1]) for a in resource_id_list])

        return obj_set, resource_id

    def _validate_request(self, request):
        if not isinstance(request, Request):
            raise AuthInvalidRequest("request should be instance of iam.auth.models.Request")

        request.validate()

    def _validate_multi_action_request(self, request):
        if not isinstance(request, MultiActionRequest):
            raise AuthInvalidRequest("request should be instance of iam.auth.models.MultiActionRequest")

        request.validate()

    def _validate_resources_list(self, resources_list):
        # resources_list = [resources] = [[Resource], [Resource]]
        if not isinstance(resources_list, list):
            raise AuthInvalidParam("resources_list should be list of [iam.auth.models.Resource]")

        if not all([isinstance(r, list) for r in resources_list]):
            raise AuthInvalidParam("resources_list should be list of [iam.auth.models.Resource]")

        if not all([isinstance(i, Resource) for r in resources_list for i in r]):
            raise AuthInvalidParam("resources should be list of iam.auth.models.Resource")

    def _validate_resources_list_same_local_only(self, system, resources_list):
        # 校验, resources_list中只能是本地同一类的资源
        resource_types = {}
        for rs in resources_list:
            for r in rs:
                if system != r.system:
                    raise AuthInvalidParam(
                        "resources_list not support auth for resource belong other system: %s" % r.system
                    )
                resource_types[r.type] = 1

        if len(resource_types) != 1:
            raise AuthInvalidParam(
                "resources_list should all with the same resource_type, but got %s" % resource_types.keys()
            )

    @cached(cache=TTLCache(maxsize=1024, ttl=10), key=hash_key)
    def is_allowed_with_cache(self, request):
        return self.is_allowed(request)

    def is_allowed(self, request):
        """
        单个资源是否有权限校验
        request中会带resource到IAM, IAM会进行两阶段计算, 即resources也会参与到计算中

        支持:
        - 本地资源 resources中只有本地资源
        - 跨系统资源依赖 resources中有本地也有远程资源 (此时resoruces一定要传, 因为需要IAM帮助获取跨系统资源)
        """
        logger.debug("calling IAM.is_allowed(request)......")

        # 1. validate
        self._validate_request(request)

        # 2. _client.policy_query
        policies = self._do_policy_query(request)

        logger.debug("the return policies: %s", policies)
        if not policies:
            logger.debug("no return policies, will return False")
            return False

        # 3. make objSet
        obj_set, _ = self._build_object_set(request.system, request.resources, only_local=True)

        # 4. eval
        allowed = self._eval_policy(policies, obj_set)
        return allowed

    def is_allowed_with_policy_cache(self, request):
        """
        单个资源是否有权限校验, 缓存查询得到的策略
        同一个subject-system-action查询到的策略会被缓存
        不同的实例使用同一份缓存的策略, 提升鉴权性能

        策略缓存1分钟
        """
        # 1. validate
        self._validate_request(request)

        # 2. request without resources
        request_without_resources = deepcopy(request)
        request_without_resources.resources = []

        # 3. query policy from iam or cache
        policies = self._do_policy_query_with_cache(request_without_resources)

        logger.debug("the return policies: %s", policies)
        if not policies:
            logger.debug("no return policies, will return False")
            return False

        # 4. make objSet
        obj_set, _ = self._build_object_set(request.system, request.resources, only_local=True)

        # 5. eval
        allowed = self._eval_policy(policies, obj_set)
        return allowed

    def batch_is_allowed(self, request, resources_list):
        """
        多个资源是否有权限校验
        request中不会带resource到IAM, IAM不会会进行两阶段计算, 直接返回system+action+subejct的所有策略
        然后逐一计算

        - 一次策略查询, 多次计算

        支持:
        - 本地资源 resources中只有本地资源
        - **不支持**跨系统资源依赖
        """
        logger.debug("calling IAM.batch_is_allowed(request, resources_list)......")

        # 1. validate
        self._validate_request(request)
        self._validate_resources_list(resources_list)
        self._validate_resources_list_same_local_only(request.system, resources_list)

        # 2. _client.policy_query
        data = request.to_dict()
        logger.debug("the request: %s", data)

        # NOTE: 不向服务端传任何resource
        result = {}
        policies = self._do_policy_query(request, with_resources=False)
        logger.debug("the return policies: %s", policies)
        if not policies:
            logger.debug("no return policies, will return False")
            for resources in resources_list:
                _, resource_id = self._build_object_set(request.system, resources, only_local=False)
                result[resource_id] = False

            return result

        expr = make_expression(policies)

        # 4. make objSet
        for resources in resources_list:
            obj_set, resource_id = self._build_object_set(request.system, resources, only_local=False)

            allowed = self._eval_expr(expr, obj_set)
            result[resource_id] = allowed

        return result

    def resource_multi_actions_allowed(self, request):
        """
        单个资源多个action是否有权限校验
        request中会带resource到IAM, IAM会进行两阶段计算, 即resources也会参与到计算中

        支持:
        - 本地资源 resources中只有本地资源
        - 跨系统资源依赖 resources中有本地也有远程资源 (此时resoruces一定要传, 因为需要IAM帮助获取跨系统资源)
        """
        logger.debug("calling IAM.resource_multi_actions_allowed(request)......")

        # 1. validate
        self._validate_multi_action_request(request)

        data = request.to_dict()
        logger.debug("the request: %s", data)
        # 2. _client.policy_query_by_actions
        action_policies = self._do_policy_query_by_actions(request)

        actions_allowed = {}
        logger.debug("the return policies: %s", action_policies)
        if not action_policies:
            logger.debug("no return policies, will reject all perms")
            for a in request.to_dict()["actions"]:
                action = a["id"]
                actions_allowed[action] = False
            return actions_allowed

        # 3. calculate perms
        obj_set, _ = self._build_object_set(request.system, request.resources, only_local=True)

        # 4. 一个策略是一个表达式, 计算一次
        for action_policy in action_policies:
            action = action_policy["action"]["id"]
            policies = action_policy["condition"]
            actions_allowed[action] = self._eval_policy(policies, obj_set)

        return actions_allowed

    def batch_resource_multi_actions_allowed(self, request, resources_list):
        """
        批量资源多个action是否有权限校验
        request中会带resource到IAM, IAM会进行两阶段计算, 即resources也会参与到计算中

        支持:
        - 本地资源 resources中只有本地资源
        - **不支持**跨系统资源依赖
        """
        logger.debug("calling IAM.batch_resource_multi_actions_allowed(request, resources_list)......")

        # 1. validate
        self._validate_multi_action_request(request)
        self._validate_resources_list(resources_list)
        self._validate_resources_list_same_local_only(request.system, resources_list)

        data = request.to_dict()
        logger.debug("the request: %s", data)

        # 2. _client.policy_query_by_actions
        # NOTE: 不向服务端传任何resource
        action_policies = self._do_policy_query_by_actions(request, with_resources=False)

        resources_actions_perms = {}

        logger.debug("the return policies: %s", action_policies)
        if not action_policies:
            logger.debug("no return policies, will reject all perms")
            for resource in resources_list:
                for a in data["actions"]:
                    action = a["id"]
                    resources_actions_perms.setdefault(resource.id, {})[action] = False
            return resources_actions_perms

        # 4. calculate perms
        for resources in resources_list:
            # NOTE: 这里假设resources里面只有一个本地资源
            obj_set, resource_id = self._build_object_set(request.system, resources, only_local=False)
            # FIXME: 未来这里会支持同一个系统的不同资源, 届时怎么表示?

            # 一个策略是一个表达式, 计算一次
            for action_policy in action_policies:
                action = action_policy["action"]["id"]
                policies = action_policy["condition"]

                resources_actions_perms.setdefault(resource_id, {})[action] = False
                resources_actions_perms[resource_id][action] = self._eval_policy(policies, obj_set)

        return resources_actions_perms

    def make_filter(self, request, converter_class=DjangoQuerySetConverter, key_mapping=None):
        logger.debug("calling IAM.make_filter(request)......")

        # 1. validate
        if not isinstance(request, Request):
            raise AuthInvalidRequest("request should be instance of iam.auth.models.Request")

        request.validate()

        # 2. _client.policy_query
        policies = self._do_policy_query(request)

        # the polices maybe none
        logger.debug("the return policies: %s", policies)
        if not policies:
            return None

        # 3. make converter
        c = converter_class(key_mapping)

        # 4. do convert and return
        converted_filters = c.convert(policies)
        logger.debug("the converted filters: %s", converted_filters)
        return converted_filters

    # TODO: add the register model apis
    def get_token(self, system):
        # bool, message, token
        return self._client.get_token(system)

    def is_basic_auth_allowed(self, system, basic_auth):
        logger.debug("calling IAM.is_basic_auth_allowed(basic_aut)......")

        if not isinstance(system, string_types):
            raise AuthInvalidParam("system should be a string")

        if not isinstance(basic_auth, string_types):
            raise AuthInvalidParam("basic_auth should be a string")

        auth = basic_auth.strip().split()
        if len(auth) != 2 or auth[0].lower() != "basic":
            logger.error("invalid basic auth format [length=%d, authorization=%s]", len(auth), auth)
            return False

        decoded_str = base64.b64decode(auth[1])

        if sys.version_info[0] >= 3:
            if not isinstance(decoded_str, str):
                decoded_str = decoded_str.decode()

        username, password = decoded_str.split(":")
        if username != "bk_iam":
            logger.error("username is not bk_iam")
            return False

        ok, message, token = self.get_token(system)
        if not ok:
            logger.error("get system token fail: %s", message)
            return False

        if password != token:
            logger.error(
                "password in basic_auth not equals to system token [password=%s***, token=%s***]",
                password[6:],
                token[6:],
            )
            return False

        return True

    def get_apply_url(self, application, bk_token=None, bk_username=None):
        if isinstance(application, dict):
            data = application
        elif isinstance(application, Application):
            # do validate
            application.validate()
            data = application.to_dict()
        else:
            raise AuthInvalidRequest("application shuld be instance of dict or iam.apply.modles.Application")

        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        # bool, message, url
        return self._client.get_apply_url(bk_token, bk_username, data)

    def grant_resource_creator_actions(self, application, bk_token=None, bk_username=None):
        if isinstance(application, dict):
            data = application
        else:
            raise AuthInvalidRequest("application should be instance of dict")

        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        # bool, message, url
        return self._client.grant_resource_creator_actions(bk_token, bk_username, data)

    def grant_resource_creator_action_attributes(self, application, bk_token=None, bk_username=None):
        if isinstance(application, dict):
            data = application
        else:
            raise AuthInvalidRequest("application should be instance of dict")

        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        # bool, message
        return self._client.grant_resource_creator_action_attributes(bk_token, bk_username, data)

    def grant_batch_resource_creator_actions(self, application, bk_token=None, bk_username=None):
        if isinstance(application, dict):
            data = application
        else:
            raise AuthInvalidRequest("application should be instance of dict")

        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        # bool, message, url
        return self._client.grant_batch_resource_creator_actions(bk_token, bk_username, data)

    def grant_or_revoke_instance_permission(self, request, bk_token=None, bk_username=None):
        if not isinstance(request, ApiAuthRequest):
            raise AuthInvalidRequest("request should be a instance of iam.auth.models.ApiAuthRequest")

        self._validate_request(request)
        data = request.to_dict()

        logger.debug("the request: %s", data)
        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        ok, message, policies = self._client.instance_authorization(bk_token, bk_username, data)
        if not ok:
            raise AuthAPIError(message)
        return policies

    def grant_or_revoke_path_permission(self, request, bk_token=None, bk_username=None):
        if not isinstance(request, ApiAuthRequest):
            raise AuthInvalidRequest("request should be a instance of iam.auth.models.ApiAuthRequest")
        self._validate_request(request)
        data = request.to_dict()

        logger.debug("the request: %s", data)
        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        ok, message, policies = self._client.path_authorization(bk_token, bk_username, data)
        if not ok:
            raise AuthAPIError(message)
        return policies

    def batch_grant_or_revoke_instance_permission(self, request, bk_token=None, bk_username=None):
        if not isinstance(request, ApiBatchAuthRequest):
            raise AuthInvalidRequest("request should be a instance of iam.auth.models.ApiBatchAuthRequest")

        self._validate_request(request)
        data = request.to_dict()

        logger.debug("the request: %s", data)
        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        ok, message, policies = self._client.batch_instance_authorization(bk_token, bk_username, data)
        if not ok:
            raise AuthAPIError(message)
        return policies

    def batch_grant_or_revoke_path_permission(self, request, bk_token=None, bk_username=None):
        if not isinstance(request, ApiBatchAuthRequest):
            raise AuthInvalidRequest("request should be a instance of iam.auth.models.ApiBatchAuthRequest")

        self._validate_request(request)
        data = request.to_dict()

        logger.debug("the request: %s", data)
        if not (bk_token or bk_username):
            raise AuthInvalidRequest("bk_token and bk_username can not both be empty")

        ok, message, policies = self._client.batch_path_authorization(bk_token, bk_username, data)
        if not ok:
            raise AuthAPIError(message)
        return policies

    def query_polices_with_action_id(self, system, data):

        logger.debug("calling IAM.query_polices_with_action_id.....")

        if not isinstance(system, string_types):
            raise AuthInvalidParam("system should be a string")

        if not isinstance(data, dict):
            raise AuthInvalidParam("data should be a dict")

        ok, message, policies = self._client.query_policies_with_action_id(system, data)
        if not ok:
            raise AuthAPIError(message)

        return policies



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import logging

import curlify
import requests

logger = logging.getLogger("iam")


def _gen_header():
    headers = {
        "Content-Type": "application/json",
    }
    return headers


def _http_request(
    method,
    url,
    headers=None,
    data=None,
    verify=False,
    cert=None,
    timeout=None,
    cookies=None,
):
    resp = requests.Response()
    try:
        if method == "GET":
            resp = requests.get(
                url=url,
                headers=headers,
                params=data,
                verify=verify,
                cert=cert,
                timeout=timeout,
                cookies=cookies,
            )
        elif method == "HEAD":
            resp = requests.head(
                url=url,
                headers=headers,
                verify=verify,
                cert=cert,
                timeout=timeout,
                cookies=cookies,
            )
        elif method == "POST":
            resp = requests.post(
                url=url,
                headers=headers,
                json=data,
                verify=verify,
                cert=cert,
                timeout=timeout,
                cookies=cookies,
            )
        elif method == "DELETE":
            resp = requests.delete(
                url=url,
                headers=headers,
                json=data,
                verify=verify,
                cert=cert,
                timeout=timeout,
                cookies=cookies,
            )
        elif method == "PUT":
            resp = requests.put(
                url=url,
                headers=headers,
                json=data,
                verify=verify,
                cert=cert,
                timeout=timeout,
                cookies=cookies,
            )
        else:
            return False, "invalid http method: %s" % method, None
    except requests.exceptions.RequestException as e:
        message = "http error! request: [method=`%s`, url=`%s`, data=`%s`] err=`%s`" % (method, url, data, str(e))
        logger.exception(message)
        return False, message, None
    else:
        request_id = resp.headers.get("X-Request-Id")

        content = resp.content if resp.content else ""
        if not logger.isEnabledFor(logging.DEBUG) and len(content) > 200:
            content = content[:200] + b"......"

        message_format = (
            "request: [method=`%s`, url=`%s`, data=`%s`] response: [status_code=`%s`, request_id=`%s`, content=`%s`]"
        )

        if resp.status_code != 200:
            message = message_format % (method, url, str(data), resp.status_code, request_id, content)
            logger.error(message)
            return False, message, None

        logger.info(message_format % (method, url, str(data), resp.status_code, request_id, content))
        return True, "ok", resp.json()
    finally:
        if resp.request is None:
            resp.request = requests.Request(method, url, headers=headers, data=data, cookies=cookies).prepare()

        if logger.isEnabledFor(logging.DEBUG):
            logger.debug(
                "the request_id: `%s`. curl: `%s`",
                resp.headers.get("X-Request-Id", ""),
                curlify.to_curl(resp.request, verify=False),
            )


def http_get(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="GET",
        url=url,
        headers=headers,
        data=data,
        verify=verify,
        cert=cert,
        timeout=timeout,
        cookies=cookies,
    )


def http_post(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="POST",
        url=url,
        headers=headers,
        data=data,
        verify=verify,
        cert=cert,
        timeout=timeout,
        cookies=cookies,
    )


def http_put(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="PUT",
        url=url,
        headers=headers,
        data=data,
        verify=verify,
        cert=cert,
        timeout=timeout,
        cookies=cookies,
    )


def http_delete(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="DELETE",
        url=url,
        headers=headers,
        data=data,
        verify=verify,
        cert=cert,
        timeout=timeout,
        cookies=cookies,
    )



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


from __future__ import unicode_literals

import argparse
import json
import os

import requests

__version__ = "1.0.0"

BK_IAM_HOST = os.getenv("BK_IAM_V3_INNER_HOST", "http://bkiam.service.consul:5001")

APP_CODE = ""
APP_SECRET = ""
data_file = ""


# =================== load json ===================


def load_data(filename):
    """
    解析JSON数据文件
    """
    data = {}
    try:
        with open(filename) as data_file:
            data = json.load(data_file)
        print("parser json data file success!")
    except Exception as error:
        print("parser json data file error: %s" % error)
        data = {}
    return data


# =================== http request ===================


def _gen_header():
    headers = {
        "Content-Type": "application/json",
    }
    return headers


def _http_request(method, url, headers=None, data=None, timeout=None, verify=False, cert=None, cookies=None):
    try:
        if method == "GET":
            resp = requests.get(
                url=url, headers=headers, params=data, timeout=timeout, verify=verify, cert=cert, cookies=cookies
            )
        elif method == "HEAD":
            resp = requests.head(url=url, headers=headers, verify=verify, cert=cert, cookies=cookies)
        elif method == "POST":
            resp = requests.post(
                url=url, headers=headers, json=data, timeout=timeout, verify=verify, cert=cert, cookies=cookies
            )
        elif method == "DELETE":
            resp = requests.delete(
                url=url, headers=headers, json=data, timeout=timeout, verify=verify, cert=cert, cookies=cookies
            )
        elif method == "PUT":
            resp = requests.put(
                url=url, headers=headers, json=data, timeout=timeout, verify=verify, cert=cert, cookies=cookies
            )
        else:
            return False, None
    except requests.exceptions.RequestException as e:
        print("http request error! method: %s, url: %s, data: %s! err=%s", method, url, data, e)
        return False, None
    else:
        if resp.status_code != 200:
            content = resp.content[:100] if resp.content else ""
            error_msg = (
                "http request fail! method: %s, url: %s, data: %s, " "response_status_code: %s, response_content: %s"
            )
            print(error_msg % (method, url, str(data), resp.status_code, content))
            return False, None

        return True, resp.json()


def http_get(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="GET", url=url, headers=headers, data=data, verify=verify, cert=cert, timeout=timeout, cookies=cookies
    )


def http_post(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="POST", url=url, headers=headers, data=data, timeout=timeout, verify=verify, cert=cert, cookies=cookies
    )


def http_put(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="PUT", url=url, headers=headers, data=data, timeout=timeout, verify=verify, cert=cert, cookies=cookies
    )


def http_delete(url, data, headers=None, verify=False, cert=None, timeout=None, cookies=None):
    if not headers:
        headers = _gen_header()
    return _http_request(
        method="DELETE", url=url, headers=headers, data=data, timeout=timeout, verify=verify, cert=cert, cookies=cookies
    )


# =================== iam func ===================


class Client(object):
    def __init__(self, app_code, app_secret, bk_iam_host):
        self.app_code = app_code
        self.app_secret = app_secret
        self.bk_iam_host = bk_iam_host
        self.system_id_set = set()
        self.resource_id_set = set()
        self.action_id_set = set()

    # 调用权限中心方法
    def _call_iam_api(self, http_func, path, data):
        headers = {"X-BK-APP-CODE": self.app_code, "X-BK-APP-SECRET": self.app_secret}
        url = "{host}{path}".format(host=self.bk_iam_host, path=path)
        ok, _data = http_func(url, data, headers=headers)
        # TODO: add debug here
        if not ok:
            message = "verify from iam server fail"
            print("_call_iam_api fail.", "error:", message)
            return False, message, None

        if _data.get("code") != 0:
            message = _data.get("message", "iam api fail")

            if not (
                http_func.__name__ == "http_get"
                and path.startswith("/api/v1/model/systems/")
                and path.endswith("/query")
                and message.startswith("not found:system(")
                and message.endswith(") not exists")
            ):
                print("_call_iam_api fail.", "method:", http_func.__name__, "path:", path, "error:", message)

            return False, message, None

        _d = _data.get("data")

        return True, "ok", _d

    # all operations
    operation_funcs = {
        "add_system": "add_system",
        "update_system": "update_system",
        "upsert_system": "upsert_system",
        "add_resource_type": "add_resource_type",
        "update_resource_type": "update_resource_type",
        "delete_resource_type": "delete_resource_type",
        "upsert_resource_type": "upsert_resource_type",
        "add_instance_selection": "add_instance_selection",
        "update_instance_selection": "update_instance_selection",
        "delete_instance_selection": "delete_instance_selection",
        "upsert_instance_selection": "upsert_instance_selection",
        "add_action": "add_action",
        "update_action": "update_action",
        "delete_action": "delete_action",
        "upsert_action": "upsert_action",
        "add_action_groups": "add_action_groups",
        "update_action_groups": "update_action_groups",
        "upsert_action_groups": "update_action_groups",
        "add_resource_creator_actions": "add_resource_creator_actions",
        "update_resource_creator_actions": "update_resource_creator_actions",
        "upsert_resource_creator_actions": "update_resource_creator_actions",
        "add_common_actions": "add_common_actions",
        "update_common_actions": "update_common_actions",
        "upsert_common_actions": "update_common_actions",
        "add_feature_shield_rules": "add_feature_shield_rules",
        "update_feature_shield_rules": "update_feature_shield_rules",
        "upsert_feature_shield_rules": "update_feature_shield_rules",
    }

    """
    index:
    - Add system
    - Update system

    - Add resource types(batch)
    - Update resource type(one by one)
    - Delete resource types(batch)

    - Add actions(batch)
    - Update action(one by one)
    - Delete actions(batch)

    - Query
    """

    # ---------- system
    def api_add_system(self, data):
        path = "/api/v1/model/systems"
        ok, message, data = self._call_iam_api(http_post, path, data)
        # if alreay exists, return true
        return ok, message

    def api_update_system(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    # ---------- resource_type

    def api_batch_add_resource_types(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/resource-types".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        # if alreay exists, return true
        return ok, message

    def api_update_resource_type(self, system_id, resource_type_id, data):
        path = "/api/v1/model/systems/{system_id}/resource-types/{resource_type_id}".format(
            system_id=system_id, resource_type_id=resource_type_id
        )
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    def api_batch_delete_resource_types(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/resource-types?check_existence=false".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_delete, path, data)
        return ok, message

    # ---------- instance_selection

    def api_batch_add_instance_selections(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/instance-selections".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        # if alreay exists, return true
        return ok, message

    def api_update_instance_selection(self, system_id, instance_selection_id, data):
        path = "/api/v1/model/systems/{system_id}/instance-selections/{instance_selection_id}".format(
            system_id=system_id, instance_selection_id=instance_selection_id
        )
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    def api_batch_delete_instance_selections(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/instance-selections?check_existence=false".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_delete, path, data)
        return ok, message

    # ---------- action

    def api_batch_add_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        # if alreay exists, return true
        return ok, message

    def api_update_action(self, system_id, action_id, data):
        path = "/api/v1/model/systems/{system_id}/actions/{action_id}".format(system_id=system_id, action_id=action_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    def api_batch_delete_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/actions?check_existence=false".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_delete, path, data)
        return ok, message

    # ---------- action_groups
    def api_add_action_groups(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/action_groups".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message

    def api_update_action_groups(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/action_groups".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    # ---------- resource_creator_actions
    def api_add_resource_creator_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/resource_creator_actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message

    def api_update_resource_creator_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/resource_creator_actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    # ---------- common_actions
    def api_add_common_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/common_actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message

    def api_update_common_actions(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/common_actions".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    # ---------- feature_shield_rules
    def api_add_feature_shield_rules(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/feature_shield_rules".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_post, path, data)
        return ok, message

    def api_update_feature_shield_rules(self, system_id, data):
        path = "/api/v1/model/systems/{system_id}/configs/feature_shield_rules".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_put, path, data)
        return ok, message

    # ---------- query

    def api_query(self, system_id):
        path = "/api/v1/model/systems/{system_id}/query".format(system_id=system_id)
        ok, message, data = self._call_iam_api(http_get, path, None)
        return ok, message, data

    # =================== operations ===================

    def add_system(self, system_id, data):
        d_system_id = data.get("id")
        if not d_system_id:
            return False, "the field `id` required"

        if system_id != d_system_id:
            return False, "json[system_id] is not equals the value of `id`"

        return self.api_add_system(data)

    def update_system(self, system_id, data):
        d_system_id = data.get("id")
        if not d_system_id:
            return False, "the field `id` required"

        if system_id != d_system_id:
            return False, "json[system_id] is not equals the value of `id`"

        return self.api_update_system(system_id, data)

    def add_resource_type(self, system_id, data):
        d_resource_type_id = data.get("id")
        if not d_resource_type_id:
            return False, "the field `id` required"

        d = [data]
        return self.api_batch_add_resource_types(system_id, d)

    def update_resource_type(self, system_id, data):
        d_resource_type_id = data.get("id")
        if not d_resource_type_id:
            return False, "the field `id` required"

        data.pop("id")
        return self.api_update_resource_type(system_id, d_resource_type_id, data)

    def delete_resource_type(self, system_id, data):
        d_resource_type_id = data.get("id")
        if not d_resource_type_id:
            return False, "the field `id` required"

        d = [{"id": d_resource_type_id}]

        return self.api_batch_delete_resource_types(system_id, d)

    def add_instance_selection(self, system_id, data):
        d_instance_selection_id = data.get("id")
        if not d_instance_selection_id:
            return False, "the field `id` required"

        d = [data]
        return self.api_batch_add_instance_selections(system_id, d)

    def update_instance_selection(self, system_id, data):
        d_instance_selection_id = data.get("id")
        if not d_instance_selection_id:
            return False, "the field `id` required"

        data.pop("id")
        return self.api_update_instance_selection(system_id, d_instance_selection_id, data)

    def delete_instance_selection(self, system_id, data):
        d_instance_selection_id = data.get("id")
        if not d_instance_selection_id:
            return False, "the field `id` required"

        d = [{"id": d_instance_selection_id}]

        return self.api_batch_delete_instance_selections(system_id, d)

    def add_action(self, system_id, data):
        d_action_id = data.get("id")
        if not d_action_id:
            return False, "the field `id` required"

        d = [data]
        return self.api_batch_add_actions(system_id, d)

    def update_action(self, system_id, data):
        d_action_id = data.get("id")
        if not d_action_id:
            return False, "the field `id` required"

        return self.api_update_action(system_id, d_action_id, data)

    def delete_action(self, system_id, data):
        d_action_id = data.get("id")
        if not d_action_id:
            return False, "the field `id` required"

        d = [{"id": d_action_id}]

        return self.api_batch_delete_actions(system_id, d)

    def add_action_groups(self, system_id, data):
        return self.api_add_action_groups(system_id, data)

    def update_action_groups(self, system_id, data):
        return self.api_update_action_groups(system_id, data)

    def add_common_actions(self, system_id, data):
        return self.api_add_common_actions(system_id, data)

    def update_common_actions(self, system_id, data):
        return self.api_update_common_actions(system_id, data)

    def add_resource_creator_actions(self, system_id, data):
        return self.api_add_resource_creator_actions(system_id, data)

    def update_resource_creator_actions(self, system_id, data):
        return self.api_update_resource_creator_actions(system_id, data)

    def add_feature_shield_rules(self, system_id, data):
        return self.api_add_feature_shield_rules(system_id, data)

    def update_feature_shield_rules(self, system_id, data):
        return self.api_update_feature_shield_rules(system_id, data)

    def upsert_system(self, system_id, data):
        if system_id not in self.system_id_set:
            return self.add_system(system_id, data)
        return self.update_system(system_id, data)

    def upsert_resource_type(self, system_id, data):
        d_resource_type_id = data.get("id")
        if not d_resource_type_id:
            return False, "the field `id` required"

        if d_resource_type_id not in self.resource_id_set:
            return self.add_resource_type(system_id, data)
        return self.update_resource_type(system_id, data)

    def upsert_instance_selection(self, system_id, data):
        d_instance_selection_id = data.get("id")
        if not d_instance_selection_id:
            return False, "the field `id` required"

        if d_instance_selection_id not in self.instance_selection_id_set:
            return self.add_instance_selection(system_id, data)
        return self.update_instance_selection(system_id, data)

    def upsert_action(self, system_id, data):
        d_action_id = data.get("id")
        if not d_action_id:
            return False, "the field `id` required"

        if d_action_id not in self.action_id_set:
            return self.add_action(system_id, data)
        return self.update_action(system_id, data)

    def query_all_models(self, system_id):
        ok, message, data = self.api_query(system_id)
        if not ok:
            # ignore the first migration do_migrate fail
            if "0001_" in data_file:
                pass
            else:
                print(
                    "[Ignore this message if do_migrate the first file 0001_*.json, "
                    "because the system is not registered yet] do api_query fail",
                    message,
                )
            return set(), set(), set(), set()

        system = data.get("base_info", {}) or {}
        resource_types = data.get("resource_types", []) or []
        actions = data.get("actions", []) or []
        instance_selections = data.get("instance_selections", []) or []

        system_ids = {system.get("id")}
        resource_type_ids = {r.get("id") for r in resource_types}
        action_ids = {a.get("id") for a in actions}
        instance_selection_ids = {r.get("id") for r in instance_selections}

        return system_ids, resource_type_ids, action_ids, instance_selection_ids

    def do_operation(self, op, system_id, data):
        if op not in self.operation_funcs:
            print("invalid operation: %s" % op)
            exit(1)

        return getattr(self, self.operation_funcs[op])(system_id, data)

    def setup_models(self, system_id_set, resource_id_set, action_id_set, instance_selection_id_set):
        self.system_id_set = system_id_set
        self.resource_id_set = resource_id_set
        self.action_id_set = action_id_set
        self.instance_selection_id_set = instance_selection_id_set


# ---------- ping


def api_ping(bk_iam_host):
    url = "{host}{path}".format(host=bk_iam_host, path="/ping")
    ok, data = http_get(url, None, timeout=5)
    return ok, data


def do_migrate(data, bk_iam_host=BK_IAM_HOST, app_code=APP_CODE, app_secret=APP_SECRET):
    system_id = data.get("system_id")
    if not system_id:
        print("invald json. [system_id] required, and should not be empty")
        return False

    operations = data.get("operations")
    if not operations:
        print("invald json. [operations] required, and should not be empty")
        return False

    print("do migrate")

    client = Client(app_code, app_secret, bk_iam_host)

    # 1. query all data of the system
    system_ids, resource_type_ids, action_ids, instance_selection_ids = client.query_all_models(system_id)

    client.setup_models(system_ids, resource_type_ids, action_ids, instance_selection_ids)

    for op in operations:
        operation = op.get("operation")
        if not operation:
            print("there got a empty `operation` in the json, will ignore and continue")
            continue
            # print("")
            # return False

        data = op.get("data")
        if not data:
            print("no `data` in the json body or the `data` is empty, operation=%s" % operation)
            return False

        op_data_id = ""
        if isinstance(data, dict):
            op_data_id = "id=%s" % data.get("id")

        ok, message = client.do_operation(operation, system_id, data)
        if not ok:
            print("execute operation [%s] %s fail, error message: %s" % (operation, message, op_data_id))
            return False
        print("execute operation [%s] %s success!" % (operation, op_data_id))

    print("end migrate")
    return True


if __name__ == "__main__":
    p = argparse.ArgumentParser()
    p.add_argument(
        "-t", action="store", dest="bk_iam_host", help="bk_iam_host, i.e: http://iam.service.consul", required=True
    )
    p.add_argument(
        "-f",
        action="store",
        dest="json_data_file",
        help="which migration file to execute, i.e: 00001_bk_cmdb_20190618100210.json",
        required=True,
    )
    p.add_argument("-a", action="store", dest="app_code", help="app code", required=True)
    p.add_argument("-s", action="store", dest="app_secret", help="app secret", required=True)
    args = p.parse_args()

    BK_IAM_HOST = args.bk_iam_host.rstrip("/")

    if not BK_IAM_HOST.startswith("http://"):
        BK_IAM_HOST = "http://%s" % BK_IAM_HOST

    data_file = args.json_data_file
    APP_CODE = args.app_code
    APP_SECRET = args.app_secret

    # test ping
    ok, _ = api_ping(BK_IAM_HOST)
    if not ok:
        print("iam service is not available: %s" % BK_IAM_HOST)
        exit(1)

    print("start migrate [%s]" % data_file)

    # 数据解析
    data = load_data(data_file)
    if not data:
        exit(1)

    ok = do_migrate(data, BK_IAM_HOST, APP_CODE, APP_SECRET)
    if not ok:
        print("do migrate [%s] fail" % data_file)
        exit(1)
    print("do migrate [%s] success!" % data_file)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import operator
from six.moves import reduce

from django.db.models import Q

from iam.eval.constants import KEYWORD_BK_IAM_PATH_FIELD_SUFFIX, OP
from iam.eval.expression import field_value_convert

from .base import Converter

"""
表达式解析

convert expression to django queryset
"""


class DjangoQuerySetConverter(Converter):
    def __init__(self, key_mapping=None, value_hooks=None):
        super(DjangoQuerySetConverter, self).__init__(key_mapping)

        self.value_hooks = value_hooks or {}

    def _positive(self, fmt, left, right):
        is_array = isinstance(right, (list, tuple))
        if is_array:
            kwargs_list = [{fmt.format(left): r} for r in right]
            return reduce(operator.or_, [Q(**kw) for kw in kwargs_list])

        kwargs = {fmt.format(left): right}
        return Q(**kwargs)

    def _negative(self, fmt, left, right):
        is_array = isinstance(right, (list, tuple))
        if is_array:
            kwargs_list = [{fmt.format(left): r} for r in right]
            return reduce(operator.and_, [~Q(**kw) for kw in kwargs_list])

        kwargs = {fmt.format(left): right}
        return ~Q(**kwargs)

    def _eq(self, left, right):
        return self._positive("{}", left, right)

    def _not_eq(self, left, right):
        return self._negative("{}", left, right)

    def _in(self, left, right):
        kwargs = {"{}__in".format(left): right}
        return Q(**kwargs)

    def _not_in(self, left, right):
        kwargs = {"{}__in".format(left): right}
        return ~Q(**kwargs)

    def _contains(self, left, right):
        return self._positive("{}__contains", left, right)

    def _not_contains(self, left, right):
        return self._negative("{}__contains", left, right)

    def _starts_with(self, left, right):
        return self._positive("{}__startswith", left, right)

    def _not_starts_with(self, left, right):
        return self._negative("{}__startswith", left, right)

    def _ends_with(self, left, right):
        return self._positive("{}__endswith", left, right)

    def _not_ends_with(self, left, right):
        return self._negative("{}__endswith", left, right)

    def _lt(self, left, right):
        return self._positive("{}__lt", left, right)

    def _lte(self, left, right):
        return self._positive("{}__lte", left, right)

    def _gt(self, left, right):
        return self._positive("{}__gt", left, right)

    def _gte(self, left, right):
        return self._positive("{}__gte", left, right)

    def _any(self, left, right):
        # https://stackoverflow.com/questions/33517468/always-true-q-object
        # ~Q(pk__in=[]) => not in? should check the sql/performance

        # NOTE: We think the pk is not null in mysql schema here
        return ~Q(pk=None)

    def _and(self, content):
        # print("in _and we got", [self.convert(c) for c in content])
        return reduce(operator.and_, [self.convert(c) for c in content])

    def _or(self, content):
        # print("in _or we got", [self.convert(c) for c in content])
        return reduce(operator.or_, [self.convert(c) for c in content])

    def operator_map(self, operator, field, value):
        return None

    def convert(self, data):
        op = data["op"]

        if op == OP.AND:
            return self._and(data["content"])
        elif op == OP.OR:
            return self._or(data["content"])

        value = data["value"]
        field = data["field"]

        op_func = self.operator_map(op, field, value)

        if not op_func:
            op_func = {
                OP.EQ: self._eq,
                OP.NOT_EQ: self._not_eq,
                OP.IN: self._in,
                OP.NOT_IN: self._not_in,
                OP.CONTAINS: self._contains,
                OP.NOT_CONTAINS: self._not_contains,
                OP.STARTS_WITH: self._starts_with,
                OP.NOT_STARTS_WITH: self._not_starts_with,
                OP.ENDS_WITH: self._ends_with,
                OP.NOT_ENDS_WITH: self._not_ends_with,
                OP.LT: self._lt,
                OP.LTE: self._lte,
                OP.GT: self._gt,
                OP.GTE: self._gte,
                OP.ANY: self._any,
            }.get(op)

        if op_func is None:
            raise ValueError("invalid op %s" % op)

        # 权限中心保留字预处理
        field, value = field_value_convert(op, field, value)

        # key mapping
        if self.key_mapping and field in self.key_mapping:
            field = self.key_mapping.get(field)
        # value hooks
        if (field in self.value_hooks) and callable(self.value_hooks[field]):
            value = self.value_hooks[field](value)

        return op_func(field, value)


class PathEqDjangoQuerySetConverter(DjangoQuerySetConverter):
    def operator_map(self, operator, field, value):
        if field.endswith(KEYWORD_BK_IAM_PATH_FIELD_SUFFIX) and operator == OP.STARTS_WITH:
            return self._eq



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


"""
first version model

for python3, the dataclass is a better solution, but currently we need to support both python2/python3

options:
- https://github.com/keleshev/schema
- https://github.com/Julian/jsonschema
"""


class SystemProviderConfig(object):
    def __init__(self, host, auth):
        self.host = host
        self.auth = auth


class System(object):
    def __init__(self, id, name, name_en, description, description_en, clients, provider_config):
        self.id = id
        self.name = name
        self.name_en = name_en
        self.description = description
        self.description_en = description_en
        self.clients = clients
        self.provider_config = provider_config


class ReferenceResourceType(object):
    def __init__(self, system_id, id):
        self.system_id = system_id
        self.id = id


class ResourceProviderConfig(object):
    def __init__(self, path):
        self.path = path


class ResourceType(object):
    def __init__(self, id, name, name_en, description, description_en, parents, provider_config, version):
        self.id = id
        self.name = name
        self.name_en = name_en
        self.description = description
        self.description_en = description_en

        self.parents = parents
        self.provider_config = provider_config
        self.version = version


class InstanceSelection(object):
    def __init__(self, name, name_en, resource_type_chain):
        self.name = name
        self.name_en = name_en
        # [ReferenceResourceType]
        self.resource_type_chain = resource_type_chain


class RelatedResourceType(object):
    def __init__(self, system_id, id, name_alias, name_alias_en, scope, selection_mode, instance_selection):
        self.system_id = system_id
        self.id = id
        self.name_alias = name_alias
        self.name_alias_en = name_alias_en
        self.scope = scope
        self.selection_mode = selection_mode
        self.instance_selection = instance_selection


class Action(object):
    def __init__(
        self, id, name, name_en, description, description_en, type, related_resource_types, related_actions, version
    ):
        self.id = id
        self.name = name
        self.name_en = name_en
        self.description = description
        self.description_en = description_en

        self.type = type
        self.related_resource_types = related_resource_types
        self.related_actions = related_actions
        self.version = version


class ActionTopology(object):
    def __init__(self, id, sub_actions, related_actions):
        self.id = id
        # [ActionTopology]
        self.sub_actions = sub_actions
        # [str, str]
        self.related_actions = related_actions



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import six

from .constants import KEYWORD_BK_IAM_PATH_FIELD_SUFFIX, OP
from .operators import BINARY_OPERATORS, AndOperator, OrOperator


def _parse_bk_iam_path(value):
    new_value = value
    # /biz,1/set,*/ -> /biz,1/set,
    # resource._bk_iam_path_ startswith /biz,1/set,
    if isinstance(value, six.string_types):
        if value.endswith(",*/"):
            new_value = value[:-2]
    elif isinstance(value, (list, tuple)):
        striped_value = []
        for v in value:
            if isinstance(v, six.string_types) and v.endswith(",*/"):
                striped_value.append(v[:-2])
            else:
                striped_value.append(v)

        new_value = striped_value
    else:
        pass

    return new_value


def field_value_convert(operator, field, value):
    if operator == OP.STARTS_WITH and field.endswith(KEYWORD_BK_IAM_PATH_FIELD_SUFFIX):
        value = _parse_bk_iam_path(value)
        return field, value
    # do nothing
    return field, value


def make_expression(data):
    op = data["op"]

    if op == OP.AND:
        return AndOperator([make_expression(c) for c in data["content"]])
    elif op == OP.OR:
        return OrOperator([make_expression(c) for c in data["content"]])

    if op not in BINARY_OPERATORS:
        raise ValueError("operator %s not supported" % op)

    operator = BINARY_OPERATORS[op]

    field = data["field"]
    value = data["value"]

    field, value = field_value_convert(op, field, value)

    return operator(field, value)



import os

import setuptools

here = os.path.abspath(os.path.dirname(__file__))

long_description = "bk-iam python sdk"

requires = [
    "six>=1.11.0",
    "cachetools==3.1.1",
    "requests",
    "curlify==2.2.1",
]

about = {}
with open(os.path.join(here, "iam", "__version__.py"), "r") as f:
    exec(f.read(), about)

setuptools.setup(
    name="bk-iam",
    version=about["__version__"],
    author="TencentBlueKing",
    author_email="contactus_bk@tencent.com",
    description="bk-iam python sdk",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/TencentBlueKing/iam-python-sdk",
    project_urls={
        "Bug Tracker": "https://github.com/TencentBlueKing/iam-python-sdk/issues",
    },
    packages=setuptools.find_packages(),
    install_requires=requires,
    classifiers=[
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, !=3.4.*, !=3.5.*, <4",
)



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


from collections import OrderedDict

from . import meta


def gen_perms_apply_data(system, subject, action_to_resources_list):
    """
    根据传入的参数生成无权限交互协议数据

    action_to_resources_list 应该参照以下格式:

    [
        {
            "action": Action,
            "resources_list": [[resource1, resource2], [resource1, resource2]]
        },
        ...
    ]

    单个 action 中对应的 resources_list 必须是同类型的 Resource

    """
    data = {
        "system_id": system,
        "system_name": meta.get_system_name(system),
    }

    actions = []
    for atr in action_to_resources_list:
        action_obj = atr["action"]
        resources_list = atr["resources_list"]
        action = {
            "id": action_obj.id,
            "name": meta.get_action_name(system, action_obj.id),
        }

        # 1. aggregate resources by system and type
        system_resources_list = OrderedDict({})
        for resources in resources_list:
            system_resources = OrderedDict({})

            # 1. assemble system_resources e.g. {"system1": [r1, r2], "system2": [r3]}
            for resource in resources:
                system_resources.setdefault(resource.system, []).append(resource)

            # 2. append to system_resources_list e.g.g {"system1": [[r1, r2]], "system2": [[r3]]}
            for system_id, resources in system_resources.items():
                system_resources_list.setdefault(system_id, []).append(resources)

        related_resource_types = []
        for system_id, resources_list in system_resources_list.items():
            # get resource type from last resource in resources
            a_resource = resources_list[0][-1]
            resource_types = {
                "system_id": system_id,
                "system_name": meta.get_system_name(system_id),
                "type": a_resource.type,
                "type_name": meta.get_resource_name(system_id, a_resource.type),
            }
            instances = []

            for resources in resources_list:
                for resource in resources:
                    inst_item = []
                    topo_path = None

                    if resource.attribute:
                        topo_path = resource.attribute.get("_bk_iam_path_")

                    if topo_path:
                        for part in topo_path[1:-1].split("/"):
                            rtype, rid = part.split(",")
                            inst_item.append(
                                {
                                    "type": rtype,
                                    "type_name": meta.get_resource_name(system_id, rtype),
                                    "id": rid,
                                    "name": part,
                                }
                            )

                    inst_item.append(
                        {
                            "type": resource.type,
                            "type_name": meta.get_resource_name(system_id, resource.type),
                            "id": resource.id,
                            "name": resource.attribute.get("name", "") if resource.attribute else "",
                        }
                    )

                    instances.append(inst_item)

            resource_types["instances"] = instances
            related_resource_types.append(resource_types)

        action["related_resource_types"] = related_resource_types
        actions.append(action)

    data["actions"] = actions

    return data



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import abc
import logging

import six

from iam import Action, MultiActionRequest

logger = logging.getLogger("iam")


@six.add_metaclass(abc.ABCMeta)
class IAMResourceHelper(object):
    def __init__(self, iam, system, actions):
        self.iam = iam
        self.system = system
        self.actions = actions

    @abc.abstractmethod
    def get_resources(self, bundle):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_resources_id(self, bundle):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_subject_for_alter_list(self, request, data):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_environment_for_alter_list(self, request, data):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_subject_for_alter_detail(self, request, data):
        raise NotImplementedError()

    @abc.abstractmethod
    def get_environment_for_alter_detail(self, request, data):
        raise NotImplementedError()


class IAMResourceMixin(object):
    def alter_list_data_to_serialize(self, request, data):
        helper = getattr(self._meta, "iam_resource_helper", None)
        if not helper:
            return data

        # 1. collect resources
        resources_list = []
        for bundle in data["objects"]:
            resources_list.append(helper.get_resources(bundle))

        if not resources_list:
            return data

        # 2. make request
        request = MultiActionRequest(
            helper.system,
            helper.get_subject_for_alter_list(request, data),
            [Action(action) for action in helper.actions],
            [],
            helper.get_environment_for_alter_list(request, data),
        )

        resource_actions_allowed = helper.iam.batch_resource_multi_actions_allowed(request, resources_list)
        logger.debug(
            "tastypie alter_list_data_to_serialize batch_resource_multi_actions_allowed request({}) result: {}".format(
                request.to_dict(), resource_actions_allowed
            )
        )

        # 3. assemble action allowed data
        for bundle in data["objects"]:
            rid = str(helper.get_resources_id(bundle))
            bundle.data["auth_actions"] = [
                action for action, allowed in resource_actions_allowed.get(rid, {}).items() if allowed
            ]

        return data

    def alter_detail_data_to_serialize(self, request, data):
        helper = getattr(self._meta, "iam_resource_helper", None)
        if not helper:
            return data

        bundle = data

        # 1. get resources
        resources = helper.get_resources(bundle)

        # 2. make request
        request = MultiActionRequest(
            helper.system,
            helper.get_subject_for_alter_detail(request, data),
            [Action(action) for action in helper.actions],
            resources,
            helper.get_environment_for_alter_detail(request, data),
        )

        actions_allowed = helper.iam.resource_multi_actions_allowed(request)
        logger.debug(
            "tastypie alter_detail_data_to_serialize resource_multi_actions_allowed request({}) result: {}".format(
                request.to_dict(), actions_allowed
            )
        )

        # 3. assemble action allowed data
        bundle.data["auth_actions"] = [action for action, allowed in actions_allowed.items() if allowed]

        return data



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


import abc
import json

import six


@six.add_metaclass(abc.ABCMeta)
class BaseObject(object):
    __slots__ = ()

    @abc.abstractmethod
    def to_dict(self):
        pass

    @abc.abstractmethod
    def validate(self):
        pass


class Subject(BaseObject):
    __slots__ = ("type", "id")

    def __init__(self, type, id):
        self.type = type
        self.id = id

    def validate(self):
        # Type Check
        if not isinstance(self.type, six.string_types):
            raise TypeError("Subject.type should be a string")

        if not isinstance(self.id, six.string_types):
            raise TypeError("Subject.id should be a string")

        # Value Check
        if not self.type:
            raise ValueError("Subject.type should not be empty")

        if not self.id:
            raise ValueError("Subject.id should not be empty")

    def to_dict(self):
        return {"type": self.type, "id": self.id}


class Action(BaseObject):
    __slots__ = ("id",)

    def __init__(self, id):
        self.id = id

    def validate(self):
        # Type Check
        if not isinstance(self.id, six.string_types):
            raise TypeError("Action.id should be a string")

        # Value Check
        if not self.id:
            raise ValueError("Action.id should not be empty")

    def to_dict(self):
        return {"id": self.id}


class Resource(BaseObject):
    __slots__ = ("system", "type", "id", "attribute")

    def __init__(self, system, type, id, attribute):
        self.system = system
        self.type = type
        self.id = id
        # allow to be empty or none
        self.attribute = attribute

    def validate(self):
        # Type check
        if not isinstance(self.system, six.string_types):
            raise TypeError("system should be a string")

        if not isinstance(self.type, six.string_types):
            raise TypeError("type should be a string")

        if not isinstance(self.id, six.string_types):
            raise TypeError("id should be a string")

        if self.attribute and (not isinstance(self.attribute, dict)):
            raise TypeError("attribute should be a dict")

        # Value check
        if not self.system:
            raise ValueError("Resource.system should not be empty")

        if not self.type:
            raise ValueError("Resource.type should not be empty")

        if not self.id:
            raise ValueError("Resource.id should not be empty")

    def to_dict(self):
        return {
            "system": self.system,
            "type": self.type,
            "id": self.id,
            "attribute": self.attribute or {},
        }


class ApiAuthResourceWithId(Resource):
    __slots__ = ("system", "name", "type", "id")

    def __init__(self, system, type, id, name=""):
        self.system = system
        self.type = type
        self.id = id
        self.name = name

    def to_dict(self):
        return {
            "system": self.system,
            "type": self.type,
            "id": self.id,
            "name": self.name,
        }


class ApiBatchAuthResourceWithId(Resource):
    __slots__ = ("system", "type", "instances")

    def __init__(self, system, type, instances):
        self.system = system
        self.type = type
        self.instances = instances or []

    def validate(self):
        # Type check
        if not isinstance(self.system, six.string_types):
            raise TypeError("system should be a string")

        if not isinstance(self.type, six.string_types):
            raise TypeError("type should be a string")

        if not isinstance(self.instances, list):
            raise TypeError("instances should be a list")
        # Value check
        if not self.system:
            raise ValueError("ApiBatchAuthResourceWithId.system should not be empty")

        if not self.type:
            raise ValueError("ApiBatchAuthResourceWithId.type should not be empty")

    def to_dict(self):
        return {
            "system": self.system,
            "type": self.type,
            "instances": self.instances,
        }


class ApiAuthResourceWithPath(Resource):
    __slots__ = ("system", "path", "type")

    def __init__(self, system, type, path):
        self.system = system
        self.type = type
        self.path = path or []

    def validate(self):
        # Type check
        if not isinstance(self.system, six.string_types):
            raise TypeError("system should be a string")

        if not isinstance(self.type, six.string_types):
            raise TypeError("type should be a string")

        # Value check
        if not self.system:
            raise ValueError("ApiAuthResourceWithPath.system should not be empty")

        if not self.type:
            raise ValueError("ApiAuthResourceWithPath.type should not be empty")

    def to_dict(self):
        return {
            "system": self.system,
            "type": self.type,
            "path": self.path,
        }


class ApiBatchAuthResourceWithPath(Resource):
    __slots__ = ("system", "paths", "type")

    def __init__(self, system, type, paths):
        self.system = system
        self.type = type
        self.paths = paths or []

    def validate(self):
        # Type check
        if not isinstance(self.system, six.string_types):
            raise TypeError("system should be a string")

        if not isinstance(self.type, six.string_types):
            raise TypeError("type should be a string")

        if not isinstance(self.paths, list):
            raise TypeError("paths should be a list")

        # Value check
        if not self.system:
            raise ValueError("ApiBatchAuthResourceWithPath.system should not be empty")

        if not self.type:
            raise ValueError("ApiBatchAuthResourceWithPath.type should not be empty")

    def to_dict(self):
        return {
            "system": self.system,
            "type": self.type,
            "paths": self.paths,
        }


@six.add_metaclass(abc.ABCMeta)
class BaseRequest(BaseObject):
    def __init__(self, system, subject, environment):
        self.system = system
        self.subject = subject
        self.environment = environment

    def _validate_type(self):
        # Type check
        if not isinstance(self.system, six.string_types):
            raise TypeError("system should be a string")

        # subject can be empty for some query
        if self.subject and (not isinstance(self.subject, Subject)):
            raise TypeError("subject should be a instance of iam.auth.models.Subject")

        if self.environment and (not isinstance(self.environment, dict)):
            raise TypeError("environment should be a dict")

    def _validate_value(self):
        # Value check
        if not self.system:
            raise ValueError("Request.system should not be empty")

        if self.subject:
            try:
                self.subject.validate()
            except Exception as e:
                raise ValueError("Request.subject invalid: %s" % e)

    def validate(self):
        BaseRequest._validate_type(self)
        BaseRequest._validate_value(self)


class Request(BaseRequest):
    __slots__ = ("system", "subject", "action", "resources", "environment")

    def __init__(self, system, subject, action, resources, environment):
        super(Request, self).__init__(system, subject, environment)
        self.action = action
        self.resources = resources

    def _validate_type(self):
        if not isinstance(self.action, Action):
            raise TypeError("action should be a instance of iam.auth.models.Action")

        if self.resources and (not isinstance(self.resources, list)):
            raise TypeError("resources should be a list of iam.auth.models.Resource")

        if self.resources and (not all([isinstance(resource, Resource) for resource in self.resources])):
            raise TypeError("resources should be a list of iam.auth.models.Resource")

    def _validate_value(self):
        # Value check
        try:
            self.action.validate()
        except Exception as e:
            raise ValueError("Request.action invalid: %s" % e)

        # resources can be empty
        if self.resources:
            for i, r in enumerate(self.resources):
                try:
                    r.validate()
                except Exception as e:
                    raise ValueError("Request.resources[%d] invalid: %s" % (i, e))

    def validate(self):
        super(Request, self).validate()
        self._validate_type()
        self._validate_value()

    def to_dict(self):
        return {
            "system": self.system,
            "subject": self.subject.to_dict() if self.subject else {},
            "action": self.action.to_dict(),
            "resources": [r.to_dict() for r in self.resources] if self.resources else [],
            "environment": self.environment or {},
        }

    def __hash__(self):
        data = json.dumps(self.to_dict(), sort_keys=True)
        return hash(data)


class MultiActionRequest(BaseRequest):
    __slots__ = ("system", "subject", "actions", "resources", "environment")

    def __init__(self, system, subject, actions, resources, environment):
        super(MultiActionRequest, self).__init__(system, subject, environment)
        self.actions = actions
        self.resources = resources

    def _validate_type(self):
        if not isinstance(self.actions, list):
            raise TypeError("actions should be a list")

        if not all([isinstance(action, Action) for action in self.actions]):
            raise TypeError("actions should be a list of iam.auth.models.Action")

        if self.resources and (not isinstance(self.resources, list)):
            raise TypeError("resources should be a list")

        if not all([isinstance(resource, Resource) for resource in self.resources]):
            raise TypeError("resources should be a list of iam.auth.models.Resource")

    def _validate_value(self):
        # Value check
        if self.actions:
            for i, a in enumerate(self.actions):
                try:
                    a.validate()
                except Exception as e:
                    raise ValueError("Request.actions[%d] invalid: %s" % (i, e))

        # resources can be empty
        if self.resources:
            for i, r in enumerate(self.resources):
                try:
                    r.validate()
                except Exception as e:
                    raise ValueError("Request.resources[%d] invalid: %s" % (i, e))

    def validate(self):
        super(MultiActionRequest, self).validate()
        self._validate_type()
        self._validate_value()

    def to_dict(self):
        return {
            "system": self.system,
            "subject": self.subject.to_dict() if self.subject else {},
            "actions": [a.to_dict() for a in self.actions] if self.actions else [],
            "resources": [r.to_dict() for r in self.resources] if self.resources else [],
            "environment": self.environment or {},
        }


class ApiAuthRequest(Request):
    __slots__ = ("system", "subject", "action", "resources", "environment", "operate")

    def __init__(self, system, subject, action, resources, environment, operate, asynchronous=None):
        super(Request, self).__init__(system, subject, environment)
        self.action = action
        self.resources = resources
        self.operate = operate
        self.asynchronous = asynchronous

    def _validate_type(self):
        super(ApiAuthRequest, self)._validate_type()
        if not isinstance(self.operate, six.string_types):
            raise TypeError("operate should be a string")

    def _validate_value(self):
        # Value check
        try:
            self.action.validate()
        except Exception as e:
            raise ValueError("ApiAuthRequest.action invalid: %s" % e)
        if not self.operate:
            raise ValueError("ApiAuthRequest.operate should not be empty")

    def to_dict(self):
        request_dict = {
            "system": self.system,
            "subject": self.subject.to_dict() if self.subject else {},
            "action": self.action.to_dict(),
            "resources": [r.to_dict() for r in self.resources] if self.resources else [],
            "environment": self.environment or {},
            "operate": self.operate,
        }

        if self.asynchronous is not None:
            request_dict["asynchronous"] = self.asynchronous

        return request_dict


class ApiBatchAuthRequest(Request):
    __slots__ = ("system", "subject", "actions", "resources", "operate")

    def __init__(self, system, subject, actions, resources, operate, asynchronous=None):
        self.system = system
        self.subject = subject
        self.actions = actions
        self.resources = resources
        self.operate = operate
        self.asynchronous = asynchronous

    def _validate_type(self):
        if not isinstance(self.operate, six.string_types):
            raise TypeError("operate should be a string")
        if not isinstance(self.actions, list):
            raise TypeError("actions should be a list")

    def _validate_value(self):
        # Value check
        for action in self.actions:
            try:
                action.validate()
            except Exception as e:
                raise ValueError("ApiBatchAuthRequest.action invalid: %s" % e)
        if not self.operate:
            raise ValueError("ApiBatchAuthRequest.operate should not be empty")

    def validate(self):
        self._validate_type()
        self._validate_value()

    def to_dict(self):
        request_dict = {
            "system": self.system,
            "subject": self.subject.to_dict() if self.subject else {},
            "actions": [action.to_dict() for action in self.actions] if self.actions else [],
            "resources": [r.to_dict() for r in self.resources] if self.resources else [],
            "operate": self.operate,
        }

        if self.asynchronous is not None:
            request_dict["asynchronous"] = self.asynchronous

        return request_dict



# -*- coding: utf-8 -*-
"""
TencentBlueKing is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS
Community Edition) available.
Copyright (C) 2017-2018 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""


class DictObject(object):
    def __init__(self, obj):
        self.obj = obj
        self.is_dict = isinstance(obj, dict)

    def __getattr__(self, key):
        if self.is_dict:
            return self.obj.get(key)
        else:
            if hasattr(self.obj, key):
                return getattr(self.obj, key)
            return None

    def __setattr__(self, key, value):
        if key in ("obj", "is_dict"):
            super(DictObject, self).__setattr__(key, value)
            return

        raise AttributeError("read only object")


class ObjectSet(object):
    def __init__(self):
        self._data = {}

    def add_object(self, _type, obj):
        self._data[_type] = obj

    def get_object(self, _type):
        return self._data.get(_type)

    def has_object(self, _type):
        return _type in self._data

    def del_object(self, _type):
        if _type in self._data:
            del self._data[_type]

    def get(self, key):
        parts = key.split(".")
        if len(parts) != 2:
            return None

        _type, attribute_name = parts

        has_object = self.has_object(_type)
        if not has_object:
            return None

        obj = self.get_object(_type)
        if isinstance(obj, dict):
            return obj[attribute_name]

        return getattr(obj, attribute_name)
